from __future__ import annotations

import copy
import enum
import json
import logging
import time
import typing as t
import uuid
import warnings
from collections import Counter
from datetime import datetime
from functools import partial
from pathlib import Path

from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    create_model,
    root_validator,
    validator,
)

from hrflow_connectors.core import backend
from hrflow_connectors.core.templates import Templates
from hrflow_connectors.core.warehouse import ReadMode, Warehouse

HRFLOW_CONNECTORS_RAW_GITHUB_CONTENT_BASE = (
    "https://raw.githubusercontent.com/Riminder/hrflow-connectors"
)
CONNECTORS_DIRECTORY = Path(__file__).parent.parent / "connectors"
KB = 1024
MAX_LOGO_SIZE_BYTES = 100 * KB
MAX_LOGO_PIXEL = 150
MIN_LOGO_PIXEL = 34
logger = logging.getLogger(__name__)


class ConnectorActionAdapter(logging.LoggerAdapter):
    def process(self, msg: str, kwargs: t.Dict) -> t.Tuple[str, t.Dict]:
        tags = [
            "[{}={}]".format(tag["name"], tag["value"])
            for tag in self.extra["log_tags"]
        ]
        return (
            "{}: {}".format(
                "".join(tags),
                msg,
            ),
            kwargs,
        )


class Event(str, enum.Enum):
    read_success = "read_success"
    read_failure = "read_failure"
    format_failure = "format_failure"
    logics_discard = "logics_discard"
    logics_failure = "logics_failure"
    write_failure = "write_failure"
    callback_failure = "callback_failure"
    callback_executed = "callback_executed"
    item_to_read_from_failure = "item_to_read_from_failure"

    @classmethod
    def empty_counter(cls) -> t.Counter["Event"]:
        return Counter({event: 0 for event in cls})


class Reason(str, enum.Enum):
    item_to_read_from_failure = "item_to_read_from_failure"
    origin_does_not_support_incremental = "origin_does_not_support_incremental"
    backend_not_configured_in_incremental_mode = (
        "backend_not_configured_in_incremental_mode"
    )
    workflow_id_not_found = "workflow_id_not_found"
    event_parsing_failure = "event_parsing_failure"
    bad_action_parameters = "bad_action_parameters"
    bad_origin_parameters = "bad_origin_parameters"
    bad_target_parameters = "bad_target_parameters"
    format_failure = "format_failure"
    logics_failure = "logics_failure"
    read_failure = "read_failure"
    write_failure = "write_failure"
    none = ""


class Status(str, enum.Enum):
    success = "success"
    success_with_failures = "success_with_failures"
    fatal = "fatal"


class ActionInitError(BaseModel):
    data: t.Dict
    reason: Reason


class RunResult(BaseModel):
    status: Status
    reason: Reason = Reason.none
    events: t.Counter[Event] = Field(default_factory=Event.empty_counter)
    read_from: t.Optional[str] = None

    @classmethod
    def from_events(cls, events: t.Counter[Event]) -> "RunResult":
        read_success = events[Event.read_success]
        read_failures = events[Event.read_failure]
        if read_success == 0 and read_failures == 0:
            return cls(status=Status.success, events=events)
        elif read_success == 0 and read_failures > 0:
            return cls(
                status=Status.fatal,
                reason=Reason.read_failure,
                events=events,
            )

        format_failures = events[Event.format_failure]
        if format_failures == read_success:
            return cls(
                status=Status.fatal,
                reason=Reason.format_failure,
                events=events,
            )

        logics_failures = events[Event.logics_failure]
        if logics_failures == read_success - format_failures:
            return cls(
                status=Status.fatal,
                reason=Reason.logics_failure,
                events=events,
            )

        logics_discard = events[Event.logics_discard]
        write_failure = events[Event.write_failure]
        if (
            write_failure
            == read_success - format_failures - logics_discard - logics_failures
        ) and write_failure > 0:
            return cls(
                status=Status.fatal,
                reason=Reason.write_failure,
                events=events,
            )

        success_with_failures = any(
            events[event] > 0
            for event in [
                Event.read_failure,
                Event.format_failure,
                Event.logics_failure,
                Event.write_failure,
                Event.callback_failure,
            ]
        )
        if success_with_failures:
            return cls(status=Status.success_with_failures, events=events)
        return cls(status=Status.success, events=events)


LogicFunctionType = t.Callable[[t.Dict], t.Union[t.Dict, None]]
# Different versions of Python produce different string
# reprensentations for t.Union[t.Any, None] to avoid
# inconsistencies in manifest this is hardcoded belo
LogicFunctionTypeStr = "typing.Callable[[typing.Dict], typing.Optional[typing.Dict]]"
LogicsTemplate = """
import typing as t

def logic_1(item: t.Dict) -> t.Union[t.Dict, None]:
    return None

def logic_2(item: t.Dict) -> t.Uniont[t.Dict, None]:
    return None

logics = [logic_1, logic_2]
"""
LogicsDescription = "List of logic functions"
FormatFunctionType = t.Callable[[t.Dict], t.Dict]
FormatTemplate = """
import typing as t

def format(item: t.Dict) -> t.Dict:
    return item
"""
FormatDescription = "Formatting function"
EventParserFunctionType = t.Callable[[t.Dict], t.Dict]
EventParserTemplate = """
import typing as t

def event_parser(event: t.Dict) -> t.Dict:
    parsed = dict()
    parsed["user_id"] = event["email"]
    parsed["thread_id"] = event["subscription_id"]
    return parsed
"""
EventParserDescription = "Event parsing function"
EventParserExtra = dict(skip_from_docs=True)


class BaseActionParameters(BaseModel):
    logics: t.List[LogicFunctionType] = Field(
        default_factory=list, description=LogicsDescription
    )
    format: FormatFunctionType = Field(lambda x: x, description=FormatDescription)
    event_parser: t.Optional[EventParserFunctionType] = Field(
        None, description=EventParserDescription, **EventParserExtra
    )
    read_mode: ReadMode = Field(
        ReadMode.sync,
        description=(
            "If 'incremental' then `read_from` of the last run is given to Origin"
            " Warehouse during read. **The actual behavior depends on implementation of"
            " read**. In 'sync' mode `read_from` is neither fetched nor given to Origin"
            " Warehouse during read."
        ),
    )

    class Config:
        extra = "forbid"

        @staticmethod
        def schema_extra(
            schema: t.Dict[str, t.Any], model: t.Type["BaseActionParameters"]
        ) -> None:
            # JSON has no equivalent for Callable type which is used for
            # logics, format and event_parser. Thus we hardcode properties here
            schema["properties"]["logics"] = {
                "title": "logics",
                "description": (
                    "List of logic functions. Each function should have"
                    " the following signature {}. The final list should be exposed "
                    "in a variable named 'logics'.".format(LogicFunctionTypeStr)
                ),
                "template": LogicsTemplate,
                "type": "code_editor",
            }

            schema["properties"]["format"] = {
                "title": "format",
                "description": (
                    "Formatting function. You should expose a function"
                    " named 'format' with following signature {}".format(
                        FormatFunctionType
                    )
                ),
                "template": FormatTemplate,
                "type": "code_editor",
            }

            schema["properties"]["event_parser"] = {
                "title": "event_parser",
                "description": (
                    "Event parsing function for **CATCH** integrations. You should"
                    " expose a function named 'event_parser' with following"
                    " signature {}".format(EventParserFunctionType)
                ),
                "template": EventParserTemplate,
                "type": "code_editor",
            }

    @classmethod
    def with_defaults(
        cls,
        model_name: str,
        *,
        format: t.Optional[FormatFunctionType] = None,
        event_parser: t.Optional[EventParserFunctionType] = None,
    ) -> t.Type["BaseActionParameters"]:
        new_model = cls
        if format is not None:
            new_model = create_model(
                model_name,
                format=(
                    FormatFunctionType,
                    Field(format, description=FormatDescription),
                ),
                __base__=new_model,
            )
        if event_parser is not None:
            new_model = create_model(
                model_name,
                event_parser=(
                    EventParserFunctionType,
                    Field(
                        event_parser,
                        description=EventParserDescription,
                        **EventParserExtra,
                    ),
                ),
                __base__=new_model,
            )
        return new_model


class WorkflowType(str, enum.Enum):
    catch = "hook"
    pull = "schedule"


class ActionName(str, enum.Enum):
    pull_application_list = "pull_application_list"
    pull_job_list = "pull_job_list"
    pull_profile_list = "pull_profile_list"
    pull_resume_attachment_list = "pull_resume_attachment_list"
    push_profile = "push_profile"
    push_job = "push_job"
    push_profile_list = "push_profile_list"
    push_job_list = "push_job_list"
    push_score_list = "push_score_list"
    catch_profile = "catch_profile"
    catch_job = "catch_job"
    # TalentSoft actions
    applicant_new = "applicant_new"
    applicant_resume_update = "applicant_resume_update"
    applicant_update = "applicant_update"


class ActionType(str, enum.Enum):
    """
    ActionType is used to distinguish between inbound and outbound actions.
    Inbound actions are used to fetch data from external sources and push
    it to HrFlow.ai.
    Outbound actions are used to fetch data from HrFlow.ai and push it
    to external sources.
    """

    inbound = "inbound"
    outbound = "outbound"


class ConnectorAction(BaseModel):
    WORKFLOW_FORMAT_PLACEHOLDER = "# << format_placeholder >>"
    WORKFLOW_LOGICS_PLACEHOLDER = "# << logics_placeholder >>"
    WORKFLOW_EVENT_PARSER_PLACEHOLDER = "# << event_parser_placeholder >>"
    ORIGIN_SETTINGS_PREFIX = "origin_"
    TARGET_SETTINGS_PREFIX = "target_"
    WORKFLOW_ID_SETTINGS_KEY = "__workflow_id"
    trigger_type: WorkflowType
    name: ActionName
    description: str
    parameters: t.Type[BaseModel]
    origin: Warehouse
    target: Warehouse
    callback: t.Optional[
        t.Callable[[BaseModel, BaseModel, t.Counter[Event], t.List[t.Dict]], None]
    ] = None
    action_type: ActionType

    @classmethod
    def based_on(
        cls: t.Type[t.Self],
        base: t.Self,
        connector_name: str,
        with_format: t.Optional[FormatFunctionType] = None,
        with_event_parser: t.Optional[EventParserFunctionType] = None,
    ) -> t.Self:
        default_format = base.parameters.__fields__["format"].default
        default_event_parser = base.parameters.__fields__["event_parser"].default
        parameters = BaseActionParameters.with_defaults(
            "{}{}".format(connector_name, base.parameters.__name__),
            format=with_format or default_format,
            event_parser=with_event_parser or default_event_parser,
        )
        return cls(
            name=base.name,
            trigger_type=base.trigger_type,
            description=base.description,
            parameters=parameters,
            origin=base.origin,
            target=base.target,
            callback=base.callback,
            action_type=base.action_type,
        )

    @validator("origin", pre=False)
    def origin_is_readable(cls, origin):
        if origin.is_readable is False:
            raise ValueError("Origin warehouse is not readable")
        return origin

    @validator("target", pre=False)
    def target_is_writable(cls, target):
        if target.is_writable is False:
            raise ValueError("Target warehouse is not writable")
        return target

    @validator("name", pre=False)
    def name_is_coherent_with_trigger_type(cls, v, values, **kwargs):
        if (
            v
            in [
                ActionName.pull_application_list,
                ActionName.pull_job_list,
                ActionName.pull_profile_list,
            ]
            and values["trigger_type"] != WorkflowType.pull
        ):
            raise ValueError(
                "`pull_application_list`, `pull_job_list` and `pull_profile_list`"
                " are only available for"
                " trigger_type={}".format(WorkflowType.pull)
            )
        return v

    @property
    def data_type(self) -> str:
        return self.origin.data_type.name

    def workflow_code(self, connector_name: str, workflow_type: WorkflowType) -> str:
        return Templates.get_template("workflow.py.j2").render(
            format_placeholder=self.WORKFLOW_FORMAT_PLACEHOLDER,
            logics_placeholder=self.WORKFLOW_LOGICS_PLACEHOLDER,
            event_parser_placeholder=self.WORKFLOW_EVENT_PARSER_PLACEHOLDER,
            workflow_id_settings_key=self.WORKFLOW_ID_SETTINGS_KEY,
            origin_settings_prefix=self.ORIGIN_SETTINGS_PREFIX,
            target_settings_prefix=self.TARGET_SETTINGS_PREFIX,
            connector_name=connector_name,
            action_name=self.name.value,
            type=workflow_type.name,
            origin_parameters=[
                parameter for parameter in self.origin.read.parameters.__fields__
            ],
            target_parameters=[
                parameter for parameter in self.target.write.parameters.__fields__
            ],
        )

    def run(
        self,
        connector_name: str,
        workflow_id: str,
        action_parameters: t.Dict,
        origin_parameters: t.Dict,
        target_parameters: t.Dict,
        init_error: t.Optional[ActionInitError] = None,
    ) -> RunResult:
        action_id = uuid.uuid4()
        started_at = datetime.utcnow()
        adapter = ConnectorActionAdapter(
            logger,
            dict(
                log_tags=[
                    dict(name="started_at", value=started_at.isoformat()),
                    dict(name="connector", value=connector_name),
                    dict(name="action_name", value=self.name),
                    dict(name="workflow_id", value=workflow_id),
                    dict(name="action_id", value=action_id),
                ]
            ),
        )

        if init_error is not None:
            adapter.error(
                "Failed to parse event with reason={} data={}".format(
                    repr(init_error.reason), init_error.data
                )
            )
            return RunResult(
                status=Status.fatal,
                reason=init_error.reason,
            )

        adapter.info("Starting Action")
        try:
            parameters = self.parameters(**action_parameters)
        except ValidationError as e:
            adapter.warning(
                "Failed to parse action_parameters with errors={}".format(e.errors())
            )
            return RunResult(status=Status.fatal, reason=Reason.bad_action_parameters)

        try:
            origin_parameters = self.origin.read.parameters(**origin_parameters)
        except ValidationError as e:
            adapter.warning(
                "Failed to parse origin_parameters with errors={}".format(e.errors())
            )
            return RunResult(status=Status.fatal, reason=Reason.bad_origin_parameters)

        try:
            target_parameters = self.target.write.parameters(**target_parameters)
        except ValidationError as e:
            adapter.warning(
                "Failed to parse target_parameters with errors={}".format(e.errors())
            )
            return RunResult(status=Status.fatal, reason=Reason.bad_target_parameters)

        if parameters.read_mode is ReadMode.incremental:
            if self.origin.supports_incremental is False:
                adapter.warning(
                    "Origin warehouse {} does not support '{}' read mode".format(
                        self.origin.name, ReadMode.incremental.value
                    )
                )
                return RunResult(
                    status=Status.fatal,
                    reason=Reason.origin_does_not_support_incremental,
                )

            if backend.is_configured is False:
                adapter.warning(
                    "For '{}' read_mode backend must be configured".format(
                        ReadMode.incremental.value
                    )
                )
                return RunResult(
                    status=Status.fatal,
                    reason=Reason.backend_not_configured_in_incremental_mode,
                )

        read_from = None
        if parameters.read_mode is ReadMode.incremental:
            adapter.info(
                "Read mode is '{}' fetching last run results".format(
                    ReadMode.incremental.value
                )
            )
            last_results = backend.store.load(key=workflow_id, parse_as=RunResult)
            read_from = last_results.read_from if last_results is not None else None

        events = Event.empty_counter()

        read_started_at = time.time()
        adapter.info(
            "Starting to read from warehouse={} with mode={} read_from={} parameters={}"
            .format(
                self.origin.name,
                parameters.read_mode,
                read_from,
                origin_parameters,
            )
        )
        origin_adapter = ConnectorActionAdapter(
            logger,
            dict(
                log_tags=adapter.extra["log_tags"]
                + [
                    dict(name="warehouse", value=self.origin.name),
                    dict(name="action", value="read"),
                ]
            ),
        )
        origin_items = []
        try:
            for item in self.origin.read(
                origin_adapter,
                origin_parameters,
                read_mode=parameters.read_mode,
                read_from=read_from,
            ):
                origin_items.append(item)
                events[Event.read_success] += 1
        except Exception as e:
            events[Event.read_failure] += 1
            adapter.exception(
                "Failed to read from warehouse={} with parameters={} error={}".format(
                    self.origin.name, origin_parameters, repr(e)
                )
            )
        if len(origin_items) == 0:
            if events[Event.read_failure] > 0:
                adapter.warning(
                    "No items fetched from origin warehoue. Aborting action after"
                    " read_failure"
                )
            return RunResult.from_events(events)

        read_finished_at = time.time()
        adapter.info(
            "Finished reading in {} from warehouse={} n_items={} read_failure={}"
            .format(
                read_finished_at - read_started_at,
                self.origin.name,
                len(origin_items),
                events[Event.read_failure] > 0,
            )
        )

        next_read_from = read_from
        if len(origin_items) > 0 and parameters.read_mode is ReadMode.incremental:
            last_item = origin_items[-1]
            try:
                next_read_from = self.origin.item_to_read_from(last_item)
            except Exception as e:
                events[Event.item_to_read_from_failure] += 1
                adapter.exception(
                    "Failed to get read_from from warehouse={} with parameters={}"
                    " item={} error={}".format(
                        self.origin.name, origin_parameters, last_item, repr(e)
                    )
                )
                return RunResult(
                    status=Status.fatal,
                    reason=Reason.item_to_read_from_failure,
                    events=events,
                )

        using_default_format = not bool(action_parameters.get("format"))
        adapter.info(
            "Starting to format origin items using {} function".format(
                "default" if using_default_format else "user defined"
            )
        )
        formatted_items = []
        for item in origin_items:
            try:
                formatted_items.append(parameters.format(item))
            except Exception as e:
                events[Event.format_failure] += 1
                adapter.exception(
                    "Failed to format origin item using {} function error={}".format(
                        "default" if using_default_format else "user defined", repr(e)
                    )
                )
        adapter.info(
            "Finished formatting origin items success={} failures={}".format(
                len(formatted_items), events[Event.format_failure]
            )
        )

        if len(formatted_items) == 0:
            adapter.warning(
                "Formatting failed for all items. Review supplied format function."
                " Aborting action."
            )
            return RunResult.from_events(events)

        if len(parameters.logics) > 0:
            adapter.info(
                "Starting to apply logic functions: "
                "n_items={} before applying logics".format(len(formatted_items))
            )
            items_to_write = []
            for item in formatted_items:
                for i, logic in enumerate(parameters.logics):
                    try:
                        item = logic(item)
                    except Exception as e:
                        adapter.exception(
                            "Failed to apply logic function number={} error={}".format(
                                i, repr(e)
                            )
                        )
                        events[Event.logics_failure] += 1
                        break
                    if item is None:
                        events[Event.logics_discard] += 1
                        break
                else:
                    items_to_write.append(item)

            if len(items_to_write) == 0:
                adapter.warning(
                    "Logics failed for all items. Review supplied logic functions."
                    " Aborting action."
                )
                return RunResult.from_events(events)
            adapter.info(
                "Finished applying logic functions: "
                "success={} discarded={} failures={}".format(
                    len(items_to_write),
                    events[Event.logics_discard],
                    events[Event.logics_failure],
                )
            )
        else:
            adapter.info("No logic functions supplied. Skipping")
            items_to_write = formatted_items

        write_started_at = time.time()
        adapter.info(
            "Starting to write to warehouse={} with parameters={} n_items={}".format(
                self.target.name, target_parameters, len(items_to_write)
            )
        )
        target_adapter = ConnectorActionAdapter(
            logger,
            dict(
                log_tags=adapter.extra["log_tags"]
                + [
                    dict(name="warehouse", value=self.target.name),
                    dict(name="action", value="write"),
                ]
            ),
        )
        try:
            failed_items = self.target.write(
                target_adapter, target_parameters, items_to_write
            )
            events[Event.write_failure] += len(failed_items)
        except Exception as e:
            adapter.exception(
                "Failed to write to warehouse={} with parameters={} error={}".format(
                    self.target.name, target_parameters, repr(e)
                )
            )
            events[Event.write_failure] += len(items_to_write)
            return RunResult(
                status=Status.fatal,
                reason=Reason.write_failure,
                events=events,
            )
        write_finished_at = time.time()
        adapter.info(
            "Finished writing in {} to warehouse={} success={} failures={}".format(
                write_finished_at - write_started_at,
                self.target.name,
                len(items_to_write) - events[Event.write_failure],
                events[Event.write_failure],
            )
        )

        if self.callback is not None:
            adapter.info("Calling callback function")
            try:
                self.callback(
                    origin_parameters, target_parameters, events, items_to_write
                )
            except Exception as e:
                events[Event.callback_failure] += 1
                adapter.exception(
                    "Failed to run callback with error={}".format(repr(e))
                )
            finally:
                events[Event.callback_executed] += 1

        results = RunResult.from_events(events)
        results.read_from = next_read_from
        if backend.is_configured:
            adapter.info("Saving run results in {} backend".format(backend.store.name))
            backend.store.save(key=workflow_id, data=results)

        adapter.info("Finished action")
        return results


class ParametersOverride(BaseModel):
    name: ActionName
    format: t.Optional[FormatFunctionType] = None
    event_parser: t.Optional[EventParserFunctionType] = None

    @root_validator
    def not_empty(cls, values):
        if values.get("format") is None and values.get("event_parser") is None:
            raise ValueError("One of `format` or `event_parser` should not be None")
        return values


class ConnectorType(enum.Enum):
    ATS = "ATS"
    CRM = "CRM"
    HCM = "HCM"
    Automation = "Automation"
    JobBoard = "Job Board"
    Classifieds = "Classified Ads"
    Other = "Other"


class ConnectorModel(BaseModel):
    name: str
    description: str
    url: str
    type: ConnectorType
    actions: t.List[ConnectorAction]

    def logo(self, connectors_directory: Path) -> str:
        try:
            from PIL import Image, UnidentifiedImageError
        except ModuleNotFoundError:  # pragma: no cover
            raise Exception(
                "PIL is not found in current environment. Mind that you need to install"
                " the package with dev dependencies to use manifest utility"
            )
        connector_directory = connectors_directory / self.name.lower()
        if not connector_directory.is_dir():
            raise ValueError(
                "No directory found for connector {} in {}".format(
                    self.name, connector_directory
                )
            )
        logo_paths = list(connector_directory.glob("logo.*"))
        if len(logo_paths) == 0:
            raise ValueError(
                "Missing logo for connector {}. Add a logo file at {} named"
                " 'logo.(png|jpeg|...)'".format(self.name, connector_directory)
            )
        elif len(logo_paths) > 1:
            raise ValueError(
                "Found multiple logos for connector {} => {}. Only a single one should"
                " be present".format(self.name, logo_paths)
            )
        logo = logo_paths[0]
        size = logo.lstat().st_size
        if size > MAX_LOGO_SIZE_BYTES:
            raise ValueError(
                "Logo size {} KB for connector {} is above maximum limit of {} KB"
                .format(size // KB, self.name, MAX_LOGO_SIZE_BYTES // KB)
            )
        try:
            width, height = Image.open(logo).size
        except UnidentifiedImageError:
            raise ValueError(
                "Logo file for connector {} at {} doesn't seem to be a valid image"
                .format(self.name, logo)
            )

        if width != height or width > MAX_LOGO_PIXEL or width < MIN_LOGO_PIXEL:
            raise ValueError(
                "Bad logo dimensions of ({}, {}) for connector {}. Logo should have"
                " square dimensions within range {min}x{min} {max}x{max}".format(
                    width,
                    height,
                    self.name,
                    min=MIN_LOGO_PIXEL,
                    max=MAX_LOGO_PIXEL,
                )
            )
        return "{}/master/src/{}".format(
            HRFLOW_CONNECTORS_RAW_GITHUB_CONTENT_BASE,
            str(logo).split("src/")[1],
        )

    def action_by_name(self, action_name: str) -> t.Optional[ConnectorAction]:
        if "__actions_by_name" not in self.__dict__:
            self.__dict__["__actions_by_name"] = {
                action.name.value: action for action in self.actions
            }
        return self.__dict__["__actions_by_name"].get(action_name)


class Connector:
    def __init__(self, *args, **kwargs) -> None:
        self.model = ConnectorModel(*args, **kwargs)
        for action in self.model.actions:
            with_connector_name = partial(action.run, connector_name=self.model.name)
            setattr(self, action.name.value, with_connector_name)

    @classmethod
    def based_on(
        cls: t.Type[t.Self],
        base: t.Self,
        name: str,
        type: ConnectorType,
        description: str,
        url: str,
        with_parameters_override: t.Optional[t.List[ParametersOverride]] = None,
        with_actions: t.Optional[t.List[ConnectorAction]] = None,
    ) -> t.Self:
        base_actions = base.model.actions

        with_parameters_override = with_parameters_override or []
        with_actions = with_actions or []

        for parameters_override in with_parameters_override:
            base_action = next(
                (
                    action
                    for action in base_actions
                    if action.name is parameters_override.name
                ),
                None,
            )
            if base_action is None:
                raise ValueError(
                    "Base connector does not have a {} action to override".format(
                        parameters_override.name.name
                    )
                )
            duplicate = next(
                (
                    action
                    for action in with_actions
                    if action.name is parameters_override.name
                ),
                None,
            )
            if duplicate is not None:
                raise ValueError(
                    "Duplicate action name {} in `with_parameters_override` and"
                    " `with_actions`".format(parameters_override.name.name)
                )
            with_actions.append(
                ConnectorAction.based_on(
                    base=base_action,
                    connector_name=name,
                    with_format=parameters_override.format,
                    with_event_parser=parameters_override.event_parser,
                )
            )

        actions = {action.name: action for action in base_actions + with_actions}
        connector = cls(
            name=name,
            type=type,
            description=description,
            url=url,
            actions=list(actions.values()),
        )
        return connector

    def manifest(self, connectors_directory: Path) -> t.Dict:
        model = self.model
        manifest = dict(
            name=model.name,
            actions=[],
            type=model.type.value,
            logo=model.logo(connectors_directory=connectors_directory),
        )
        for action in model.actions:
            format_placeholder = action.WORKFLOW_FORMAT_PLACEHOLDER
            logics_placeholder = action.WORKFLOW_LOGICS_PLACEHOLDER
            event_parser_placeholder = action.WORKFLOW_EVENT_PARSER_PLACEHOLDER
            action_manifest = dict(
                name=action.name.value,
                action_type=action.action_type.value,
                action_parameters=copy.deepcopy(action.parameters.schema()),
                data_type=action.data_type,
                trigger_type=action.trigger_type.value,
                origin=action.origin.name,
                origin_parameters=action.origin.read.parameters.schema(),
                origin_data_schema=action.origin.data_schema.schema(),
                supports_incremental=action.origin.supports_incremental,
                target=action.target.name,
                target_parameters=action.target.write.parameters.schema(),
                target_data_schema=action.target.data_schema.schema(),
                workflow_code=action.workflow_code(
                    connector_name=model.name, workflow_type=action.trigger_type
                ),
                workflow_code_format_placeholder=format_placeholder,
                workflow_code_logics_placeholder=logics_placeholder,
                workflow_code_event_parser_placeholder=event_parser_placeholder,
                workflow_code_workflow_id_settings_key=action.WORKFLOW_ID_SETTINGS_KEY,
                workflow_code_origin_settings_prefix=action.ORIGIN_SETTINGS_PREFIX,
                workflow_code_target_settings_prefix=action.TARGET_SETTINGS_PREFIX,
            )
            if action.trigger_type is WorkflowType.pull:
                action_manifest.pop("workflow_code_event_parser_placeholder")
                action_manifest["action_parameters"]["properties"].pop("event_parser")

            manifest["actions"].append(action_manifest)
        return manifest


def hrflow_connectors_manifest(
    connectors: t.List[Connector],
    directory_path: str = ".",
    connectors_directory: Path = CONNECTORS_DIRECTORY,
) -> None:
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            message="Callable (_logics|format|event_parser) was excluded",
            category=UserWarning,
        )
        manifest = dict(
            name="HrFlow.ai Connectors",
            connectors=[
                connector.manifest(connectors_directory=connectors_directory)
                for connector in connectors
            ],
        )
    with open("{}/manifest.json".format(directory_path), "w") as f:
        f.write(json.dumps(manifest, indent=2))



import enum
import json
import typing as t
from logging import LoggerAdapter

import requests
from pydantic import Field

from hrflow_connectors.connectors.teamtailor.schema import (
    TeamtailorCandidateAttribute,
    TeamtailorJob,
)
from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)

GET_ALL_JOBS_ENDPOINT = "https://api.teamtailor.com/v1/jobs"
GET_JOB_ENDPOINT = "https://api.teamtailor.com/v1/jobs"
POST_CANDIDATE_ENDPOINT = "https://api.teamtailor.com/v1/candidates"


class RemoteStatus(str, enum.Enum):
    none = "NONE"
    hybrid = "HYBRID"
    temporary = "TEMPORARY"
    fully = "FULLY"


class JobFeed(str, enum.Enum):
    public = "PUBLIC"
    private = "PRIVATE"


class WriteProfilesParameters(ParametersModel):
    Authorization: str = Field(
        ...,
        description="Authorisation used to access Teamtailor API",
        repr=False,
        field_type=FieldType.Auth,
    )

    X_Api_Version: str = Field(
        ..., description="Dated version of the API", field_type=FieldType.Other
    )


class ReadJobsParameters(ParametersModel):
    Authorization: str = Field(
        ...,
        description="Authorisation token used to access Teamtailor API",
        repr=False,
        field_type=FieldType.Auth,
    )

    X_Api_Version: str = Field(
        ..., description="Dated version of the API", field_type=FieldType.Other
    )

    filter_status: t.Optional[RemoteStatus] = Field(
        None,
        description="Posting status of a job. One of {}".format(
            [e.value for e in RemoteStatus]
        ),
        field_type=FieldType.QueryParam,
    )
    filter_feed: t.Optional[JobFeed] = Field(
        None,
        description="Status of a job. One of {}".format([e.value for e in JobFeed]),
        field_type=FieldType.QueryParam,
    )


def enrich_location(job_id: str, headers: t.Dict) -> t.Dict:
    """
    Get_location sends a request to get the job location from its API endpoint
    """
    response = requests.request(
        "GET",
        "https://api.teamtailor.com/v1/jobs/{}/location".format(job_id),
        headers=headers,
        data={},
    )

    if not response.ok:
        raise Exception(
            response,
            message="Failed to get job location from Teamtailor.",
        )

    location = response.json().get("data")
    if location is not None:
        location_attribute = location.get("attributes")
        text = location_attribute["address"]
        city = location_attribute["city"]
        country = location_attribute["country"]
        headquarters = location_attribute["headquarters"]
        lat = location_attribute["lat"]
        lng = location_attribute["long"]
        zip = location_attribute["zip"]
        name = location_attribute["name"]
        geojson = dict(
            city=city,
            country=country,
            headquarters=headquarters,
            zip=zip,
            name=name,
        )
        location_obj = dict(text=text, geojson=geojson, lat=lat, lng=lng)
        return location_obj
    return dict(text="", geojson="", lat="", lng="")


def read(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    headers = {
        "Authorization": f"Token token={parameters.Authorization}",
        "X-Api-Version": parameters.X_Api_Version,
    }

    params = dict()
    params["filter[remote-status]"] = parameters.filter_status
    params["filter[feed]"] = parameters.filter_feed

    url = GET_ALL_JOBS_ENDPOINT
    all_jobs = []
    while True:
        response = requests.request("GET", url, headers=headers, data={}, params=params)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Teamtailor status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Teamtailor")
        response = response.json()
        jobs = response.get("data")

        all_jobs += jobs
        if response.get("links").get("next") is None:
            break
        else:
            url = response.get("links").get("next")

    adapter.info("Pulling {} jobs from Teamtailor API".format(len(all_jobs)))

    for job in all_jobs:
        full_job_response = requests.request(
            "GET",
            "{}/{}".format(GET_JOB_ENDPOINT, job.get("id")),
            headers=headers,
            data={},
        )

        if full_job_response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs details from Teamtailor job_id={}"
                " status_code={} response={}".format(
                    job["id"],
                    full_job_response.status_code,
                    full_job_response.text,
                )
            )
            raise Exception("Failed to pull jobs from Teamtailor")

        # Enrich full_job_response
        job_location = enrich_location(job.get("id"), headers)
        job_and_location = dict(job=full_job_response.json(), job_location=job_location)
        yield job_and_location


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    adapter.info("Pushing {} profiles".format(len(profiles)))
    failed_profiles = []

    for profile in profiles:
        headers = {
            "Authorization": f"Token token={parameters.Authorization}",
            "X-Api-Version": parameters.X_Api_Version,
            "Content-Type": "application/vnd.api+json",
        }
        response = requests.request(
            "POST", POST_CANDIDATE_ENDPOINT, headers=headers, data=json.dumps(profile)
        )

        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to push profile to Teamtailor, "
                " status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


TeamtailorJobWarehouse = Warehouse(
    name="Teamtailor Jobs",
    data_schema=TeamtailorJob,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read,
        endpoints=[],
    ),
)

TeamtailorProfileWarehouse = Warehouse(
    name="Teamtailor Profiles",
    data_schema=TeamtailorCandidateAttribute,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters,
        function=write,
        endpoints=[],
    ),
)



import typing as t

from pydantic import BaseModel


# Job
class Department(BaseModel):
    id: str


class JobLocation(BaseModel):
    country: t.Optional[str]
    countryCode: t.Optional[str]
    regionCode: t.Optional[str]
    region: t.Optional[str]
    city: str
    address: t.Optional[str]
    longitude: t.Optional[str]
    latitude: t.Optional[str]
    remote: t.Optional[bool]
    manual: t.Optional[bool]


class Industry(BaseModel):
    id: str


class Function(BaseModel):
    id: str


class TypeOfEmployment(BaseModel):
    id: str


class ExperienceLevel(BaseModel):
    id: t.Optional[str]


class EeoCategory(BaseModel):
    id: str


class Creator(BaseModel):
    firstName: str
    lastName: str


class Compensation(BaseModel):
    min: int
    max: int
    currency: str


class CompanyDescription(BaseModel):
    title: str
    text: str


class JobDescription(BaseModel):
    title: str
    text: str


class Qualifications(BaseModel):
    title: str
    text: str


class AdditionalInformation(BaseModel):
    title: str
    text: str


class Sections(BaseModel):
    companyDescription: CompanyDescription
    jobDescription: JobDescription
    qualifications: Qualifications
    additionalInformation: AdditionalInformation


class JobAd(BaseModel):
    sections: Sections


class SmartRecruitersJob(BaseModel):
    title: str
    refNumber: str
    createdOn: str
    updatedOn: str
    department: t.Optional[Department]
    location: JobLocation
    status: t.Optional[str]
    postingStatus: t.Optional[str]
    targetHiringDate: t.Optional[str]
    industry: t.Optional[Industry]
    function: t.Optional[Function]
    typeOfEmployment: t.Optional[TypeOfEmployment]
    experienceLevel: t.Optional[ExperienceLevel]
    eeoCategory: t.Optional[EeoCategory]
    creator: t.Optional[Creator]
    compensation: t.Optional[Compensation]
    jobAd: JobAd


# Profile
class ProfileLocation(BaseModel):
    country: str
    countryCode: str
    regionCode: str
    region: str
    city: str
    lat: int
    lng: int


class Web(BaseModel):
    skype: str
    linkedin: str
    facebook: str
    twitter: str
    website: str


class EducationItem(BaseModel):
    institution: str
    degree: str
    major: str
    current: bool
    location: str
    startDate: str
    endDate: str
    description: str


class ExperienceItem(BaseModel):
    title: str
    company: str
    current: bool
    startDate: str
    endDate: str
    location: str
    description: str


class SmartRecruitersProfile(BaseModel):
    firstName: str
    lastName: str
    email: str
    phoneNumber: str
    location: ProfileLocation
    web: Web
    tags: t.List[str]
    education: t.List[EducationItem]
    experience: t.List[ExperienceItem]



import base64
import typing as t

import requests

from hrflow_connectors.connectors.bullhorn.schemas import BullhornProfile
from hrflow_connectors.connectors.bullhorn.utils import date_format
from hrflow_connectors.connectors.bullhorn.warehouse import (
    BullhornJobWarehouse,
    BullhornProfileParsingWarehouse,
    BullhornProfileWarehouse,
)
from hrflow_connectors.connectors.hrflow.schemas import HrFlowProfile
from hrflow_connectors.connectors.hrflow.warehouse.job import HrFlowJobWarehouse
from hrflow_connectors.connectors.hrflow.warehouse.profile import (
    HrFlowProfileParsingWarehouse,
    HrFlowProfileWarehouse,
)
from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)


def to_int(elm: t.Any) -> int:
    if elm is None:
        return 0
    return int(elm)


def get_location(info: t.Dict) -> t.Dict:
    if info is not None:
        location = info.get("location")
        if location is None:
            location = dict()
        fields = location.get("fields", {})
        if fields == []:
            fields = {}
        location_dict = {
            "address1": location.get("text"),
            "city": fields.get("city"),
            "state": fields.get("country"),
            "zip": fields.get("postcode"),
        }
        return location_dict
    return None


def get_skills(data: t.Dict) -> str:
    skills = ""
    if data.get("skills") is not None:
        for i in range(len(data["skills"]) - 1):
            skills += data["skills"][i]["name"] + ", "
        skills += data["skills"][-1]["name"]
    return skills


def get_education(education_list: t.List[t.Dict]) -> t.List[t.Dict]:
    educations = []
    for hrflow_education in education_list:
        location = hrflow_education["location"]
        education = {
            "id": "0",
            "candidate": {"id": None},
            "school": hrflow_education.get("school"),
            "degree": hrflow_education.get("title"),
            "comments": hrflow_education.get("description"),
            "city": location.get("text") if location else None,
            "startDate": (
                int(
                    date_format.from_str_to_datetime(
                        hrflow_education.get("date_start")
                    ).timestamp()
                )
                if hrflow_education.get("date_start")
                else None
            ),
            "endDate": (
                int(
                    date_format.from_str_to_datetime(
                        hrflow_education.get("date_end")
                    ).timestamp()
                )
                if hrflow_education.get("date_start")
                else None
            ),
        }
        educations.append(education)
    return educations


def get_experience(experience_list: t.List[t.Dict]) -> t.List[t.Dict]:
    experience_json = []
    for hrflow_experience in experience_list:
        experience = {
            "id": "0",
            "candidate": {"id": None},
            "companyName": hrflow_experience.get("company"),
            "title": hrflow_experience.get("title"),
            "comments": hrflow_experience.get("description"),
            "startDate": (
                int(
                    date_format.from_str_to_datetime(
                        hrflow_experience.get("date_start")
                    ).timestamp()
                )
                if hrflow_experience.get("date_start")
                else None
            ),
            "endDate": (
                int(
                    date_format.from_str_to_datetime(
                        hrflow_experience.get("date_end")
                    ).timestamp()
                )
                if hrflow_experience.get("date_end")
                else None
            ),
        }
        experience_json.append(experience)
    return experience_json


def get_attachments(attachment_list: t.List[t.Dict]) -> t.List[t.Dict]:
    attachments_json = []
    for hrflow_attachment in attachment_list:
        url = hrflow_attachment["public_url"]
        response = requests.get(url)
        b64 = base64.b64encode(response.content)

        attachment = {
            "externalID": "portfolio",
            "fileContent": b64.decode(),
            "fileType": "SAMPLE",
            "name": hrflow_attachment["file_name"],
            "contentType": "text/plain",
            "description": "Resume file for candidate.",
            "type": "cover",
        }
        attachments_json.append(attachment)
    return attachments_json


def format_profile(data: HrFlowProfile) -> t.Dict:
    info = data.get("info")

    date_of_birth = None
    if info is not None and info.get("date_birth"):
        date_birth_field = info.get("date_birth")
        date_birth_timestamp = date_format.from_str_to_datetime(
            date_birth_field
        ).timestamp()
        date_of_birth = int(date_birth_timestamp)

    create_profile_body = {
        "id": data.get("reference"),
        "address": get_location(info),
        "certifications": None,
        "name": info.get("full_name") if info else None,
        "firstName": info.get("first_name") if info else None,
        "lastName": info.get("last_name") if info else None,
        "email": info.get("email") if info else None,
        "mobile": info.get("phone") if info else None,
        "dateOfBirth": date_of_birth,
        "experience": to_int(data.get("experiences_duration")),  # TODO
        "skillSet": get_skills(data) if data.get("skills") else None,
    }

    enrich_profile_education = get_education(data.get("educations"))
    enrich_profile_experience = get_experience(data.get("experiences"))
    enrich_profile_attachment = get_attachments(data.get("attachments"))
    # Four querys are needed to index a Candidate to Bullhorn
    # The querys's body are grouped in a profile_body_dict
    profile_body_dict = dict(
        create_profile_body=create_profile_body,
        enrich_profile_education=enrich_profile_education,
        enrich_profile_experience=enrich_profile_experience,
        enrich_profile_attachment=enrich_profile_attachment,
    )
    return profile_body_dict


def format_job(data: t.Dict) -> t.Dict:
    # Info
    hrflow_name = data.get("title")
    hrflow_ref = str(data.get("id"))

    # Location
    address = data.get("address")
    hrflow_fields = {
        "city": address["city"],
        "country": address["countryCode"],
        "postal_code": address["zip"],
    }
    hrflow_location = {"text": address["address1"], "fields": hrflow_fields}

    # Sections
    section_description = {
        "name": "Bullhorn_description",
        "title": "Bullhorn_description",
        "description": data["publicDescription"],
    }
    hrlflow_sections = [section_description]

    # Tags
    degree_list = data.get("degreeList")
    if degree_list:
        degree_list = ", ".join(degree_list)

    tags = []
    tags.append({"name": "durationWeeks", "value": data.get("durationWeeks")})
    tags.append({"name": "degreeList", "value": degree_list})
    tags.append({"name": "employmentType", "value": data.get("employmentType")})
    tags.append({"name": "numOpenings", "value": data.get("numOpenings")})
    tags.append({"name": "onSite", "value": data.get("onSite")})
    tags.append({"name": "salaryUnit", "value": data.get("salaryUnit")})
    tags.append({"name": "startDate", "value": data.get("startDate")})
    tags.append({"name": "status", "value": data.get("status")})
    tags.append({"name": "type", "value": data.get("type")})
    tags.append({"name": "willRelocate", "value": data.get("willRelocate")})
    tags.append({"name": "salary", "value": data.get("salary")})
    tags.append({"name": "isWorkFromHome", "value": data.get("isWorkFromHome")})
    tags.append({"name": "hoursPerWeek", "value": data.get("hoursPerWeek")})
    tags.append({"name": "hoursOfOperation", "value": data.get("hoursOfOperation")})
    tags.append({"name": "dateAdded", "value": data.get("dateAdded")})

    # Skills
    hrflow_skills = []
    skill_list = data["skillList"]
    if skill_list:
        skill_list = skill_list.split(",")
        if skill_list:
            for skill in skill_list:
                new_skill = {"name": skill, "type": "undefined", "value": None}
                hrflow_skills.append(new_skill)

    hrflow_job = {
        "name": hrflow_name,
        "reference": hrflow_ref,
        "location": hrflow_location,
        "sections": hrlflow_sections,
        "skills": hrflow_skills,
        "tags": tags,
    }

    return hrflow_job


def profile_format_parsing(data: BullhornProfile) -> t.Dict:
    profile = {}

    # Tags
    tags = []
    tags.append({"name": "dateAvailable", "value": data.get("dateAvailable")})
    tags.append({"name": "status", "value": data.get("status")})
    tags.append({"name": "employeeType", "value": data.get("employeeType")})
    tags.append(
        {"name": "activePlacements", "value": data.get("activePlacements").get("total")}
    )

    profile["reference"] = str(data.get("id"))
    profile["tags"] = tags
    profile["metadatas"] = []
    profile["created_at"] = None

    profile["resume"] = {"raw": data["cvFile"], "content_type": "application/pdf"}

    return profile


def profile_format(data: BullhornProfile) -> t.Dict:
    # Info
    first_name = data["firstName"]
    last_name = data["lastName"]
    full_name = data["name"]
    email = data["email"]
    phone = data["mobile"]
    date_birth = data["dateOfBirth"]
    gender = data["gender"]

    # Location
    location_text = data["address"]["address1"]
    location = {"text": location_text}

    info = {
        "full_name": full_name,
        "first_name": first_name,
        "last_name": last_name,
        "email": email,
        "phone": phone,
        "date_birth": date_birth,
        "location": location,
        "gender": gender,
    }

    # Tags
    tags = []
    tags.append({"name": "dateAvailable", "value": data.get("dateAvailable")})
    tags.append({"name": "status", "value": data.get("status")})
    tags.append({"name": "employeeType", "value": data.get("employeeType")})
    tags.append(
        {"name": "activePlacements", "value": data.get("activePlacements").get("total")}
    )

    # Skills
    hrflow_skills = []
    skill_list = data["skillSet"]
    if skill_list:
        skill_list = skill_list.split(",")
        if skill_list:
            for skill in skill_list:
                new_skill = {"name": skill, "type": "hard", "value": None}
                hrflow_skills.append(new_skill)

    # Education
    hrflow_education = []
    for education in data["educations"]:
        location = {"text": education["city"], "lng": None, "lat": None}
        school = education["school"]
        date_start = education["startDate"]
        date_end = education["endDate"]
        title = education["degree"]
        certifications = [education["certification"]]
        description = education["comments"]
        object_education = {
            "location": location,
            "school": school,
            "date_start": date_start,
            "date_end": date_end,
            "title": title,
            "certifications": certifications,
            "description": description,
        }
        hrflow_education.append(object_education)

    # Experience
    hrflow_experience = []
    for experience in data["workHistories"]:
        location = {"text": "", "lng": None, "lat": None}
        company = experience["companyName"]
        date_start = experience["startDate"]
        date_end = experience["endDate"]
        title = experience["title"]
        description = experience["comments"]
        title = experience["title"]
        object_experience = {
            "title": title,
            "location": location,
            "company": company,
            "date_start": date_start,
            "date_end": date_end,
            "title": title,
            "description": description,
        }
        hrflow_experience.append(object_experience)

    profile = {
        "info": info,
        "skills": hrflow_skills,
        "experiences": hrflow_experience,
        "tags": tags,
        "educations": hrflow_education,
    }
    profile["reference"] = str(data.get("id"))

    return profile


DESCRIPTION = "Transform Your Business with Bullhorn Staffing and Recruitment Software"

Bullhorn = Connector(
    name="Bullhorn",
    type=ConnectorType.ATS,
    description=DESCRIPTION,
    url="https://www.bullhorn.com/",
    actions=[
        ConnectorAction(
            name=ActionName.push_profile,
            trigger_type=WorkflowType.catch,
            description=(
                "Writes a profile from Hrflow.ai Source to Bullhorn via the API"
            ),
            parameters=BaseActionParameters.with_defaults(
                "WriteProfileActionParameters", format=format_profile
            ),
            origin=HrFlowProfileWarehouse,
            target=BullhornProfileWarehouse,
            action_type=ActionType.outbound,
        ),
        ConnectorAction(
            name=ActionName.pull_job_list,
            trigger_type=WorkflowType.pull,
            description=(
                "Retrieves jobs from Bullhorn and writes them to Hrflow.ai Board"
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadJobsActionParameters", format=format_job
            ),
            origin=BullhornJobWarehouse,
            target=HrFlowJobWarehouse,
            action_type=ActionType.inbound,
        ),
        ConnectorAction(
            name=ActionName.pull_resume_attachment_list,
            trigger_type=WorkflowType.pull,
            description=(
                "retrieves profiles attachments from Bullhorn and Parses them and sends"
                " them to Hrflow.ai source"
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadProfileActionParameters", format=profile_format_parsing
            ),
            origin=BullhornProfileParsingWarehouse,
            target=HrFlowProfileParsingWarehouse,
            action_type=ActionType.inbound,
        ),
        ConnectorAction(
            name=ActionName.pull_profile_list,
            trigger_type=WorkflowType.pull,
            description=(
                "Retrieves profiles from Bullhorn and writes them to Hrflow.ai source"
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadProfileActionParameters", format=profile_format
            ),
            origin=BullhornProfileWarehouse,
            target=HrFlowProfileWarehouse,
            action_type=ActionType.inbound,
        ),
    ],
)



import typing as t
from logging import LoggerAdapter

import requests
from pydantic import Field, HttpUrl

from hrflow_connectors.connectors.digitalrecruiters.schema import (
    DigitalRecruitersJob,
    DigitalRecruitersReadProfile,
    DigitalRecruitersWriteProfile,
)
from hrflow_connectors.core import (
    ActionEndpoints,
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)

DIGITAL_RECRUITERS_JOBS_ENDPOINT = "{url_environnement}/export/job-ads/{token}"
DIGITAL_RECRUITERS_WRITE_PROFILES_ENDPOINT = (
    "{url_environnement}/api/candidate/apply/{token}"
)
DIGITAL_RECRUITERS_READ_PROFILES_ENDPOINT = "{url_environnement}/public/v1/{endpoint}"


class ReadJobsParameters(ParametersModel):
    token: str = Field(
        ...,
        description="Digital Recruiters API token.",
        repr=False,
        field_type=FieldType.Auth,
    )
    environment_url: str = Field(
        ...,
        description="Digital Recruiters API url environnement.",
        repr=False,
        field_type=FieldType.Other,
    )


class ReadProfileParameters(ParametersModel):
    api_key: str = Field(
        ...,
        description="DigitalRecruiters API key",
        repr=False,
        field_type=FieldType.Auth,
    )
    username: str = Field(
        ...,
        description="Username for authentication",
        repr=False,
        field_type=FieldType.Auth,
    )
    password: str = Field(
        ...,
        description="Password for authentication",
        repr=False,
        field_type=FieldType.Auth,
    )
    environment_url: HttpUrl = Field(
        ...,
        description="URL environment for the API",
        repr=False,
        field_type=FieldType.Other,
    )
    jobAd: int = Field(
        None,
        description="Optional: Id of a job advertisement",
        repr=False,
        field_type=FieldType.Other,
    )
    sort: str = Field(
        None,
        description=(
            "Optional: Field to sort by (id, firstName, lastName, createdAt, updatedAt)"
        ),
        repr=False,
        field_type=FieldType.Other,
    )
    limit: int = Field(
        50,
        description="Optional: Limit the number of results returned",
        repr=False,
        field_type=FieldType.Other,
    )
    page: int = Field(
        1,
        description="Optional: Page number of results returned",
        repr=False,
        field_type=FieldType.Other,
    )


class WriteProfilesParameters(ParametersModel):
    token: str = Field(
        ...,
        description="Digital Recruiters API token.",
        repr=False,
        field_type=FieldType.Auth,
    )
    environment_url: str = Field(
        ...,
        description="Digital Recruiters API url environnement.",
        repr=False,
        field_type=FieldType.Other,
    )
    job_reference: str = Field(
        ...,
        description="reference of the job to which the candidate is applying.",
        repr=False,
        field_type=FieldType.Other,
    )
    message: t.Optional[str] = Field(
        "message du candidat",
        description="Application message.",
        repr=False,
        field_type=FieldType.Other,
    )


class TokenExpired(Exception):
    pass


def get_initial_token(params: ReadProfileParameters):
    url = f"{params.environment_url}/public/v1/users/login"
    headers = {"Content-Type": "application/json", "X-DR-API-KEY": params.api_key}
    payload = {"username": params.username, "password": params.password}

    response = requests.post(url, headers=headers, json=payload)

    if response.status_code == 200:
        return response.json()["token"], response.json()["refresh_token"]
    raise Exception(
        "Failed to get token from Digitial Recruiters API with"
        f" status_code={response.status_code} and message={response.text}"
    )


def get_refresh_token(params: ReadProfileParameters, refresh_token: str):
    url = f"{params.environment_url}/public/v1/users/refresh"

    headers = {"Content-Type": "application/json", "X-DR-API-KEY": params.api_key}

    payload = {"token": refresh_token}

    response = requests.post(url, headers=headers, json=payload)

    if response.status_code == 200:
        return response.json()["token"], response.json()["refresh_token"]
    raise Exception(
        "Failed to refresh token from Digitial Recruiters API with"
        f" status_code={response.status_code} and message={response.text}"
    )


def get_candidates(url, params: ReadProfileParameters, token):
    headers = {
        "Content-Type": "application/json",
        "X-DR-API-KEY": params.api_key,
        "Authorization": f"Bearer {token}",
    }

    params_dict = dict(
        jobAd=params.jobAd, sort=params.sort, limit=params.limit, page=params.page
    )
    response = requests.get(url, headers=headers, params=params_dict)

    if response.status_code == 200:
        data = response.json()["data"]
        return data.get("items", []), data.get("links", {}).get("next")
    elif response.status_code == 401:
        raise TokenExpired(f"Token expired. Error: {response.text}")
    raise Exception(
        "Failed to get candidates from Digitial Recruiters API with"
        f" status_code={response.status_code} and message={response.text}"
    )


def fetch_and_add_resume(candidate, token):
    cv_url = candidate.get("CV", {}).get("url")
    if cv_url:
        headers = {"Authorization": f"Bearer {token}"}
        response = requests.get(cv_url, headers=headers)

        if response.status_code == 200:
            resume_info = {
                "raw": response.content,
                "content_type": response.headers.get("Content-Type"),
            }
            candidate["resume"] = resume_info
        elif response.status_code == 401:
            raise TokenExpired(f"Token expired. Error: {response.text}")
        else:
            raise Exception(
                "Failed to get resume from Digitial Recruiters API with"
                f" status_code={response.status_code} and message={response.text}"
            )
    return candidate


def read_jobs(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    DR_JOB_URL = parameters.environment_url.rstrip("/")
    url = DIGITAL_RECRUITERS_JOBS_ENDPOINT.format(
        url_environnement=DR_JOB_URL, token=parameters.token
    )
    response = requests.get(url)
    if response.status_code != 200:
        raise Exception(
            "Failed to get jobs from Digital Recruiters API with"
            f" status_code={response.status_code} and message={response.text}"
        )
    jobs = response.json().get("ads", [])
    adapter.info(f"Found {len(jobs)} jobs in Digital Recruiters.")
    if not jobs:
        adapter.warning("No jobs found in Digital Recruiters.")
        return
    for job in jobs:
        yield job


def read_profiles(
    adapter: LoggerAdapter,
    params: ReadProfileParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    token, refresh_token = get_initial_token(params)
    next_page_link = f"{params.environment_url}/job-applications/detailed"
    while next_page_link:
        try:
            candidates, next_page_link = get_candidates(next_page_link, params, token)
        except TokenExpired:
            token, refresh_token = get_refresh_token(params, refresh_token)
            candidates, next_page_link = get_candidates(next_page_link, params, token)

        if not candidates:
            break
        for candidate in candidates:
            try:
                candidate = fetch_and_add_resume(candidate, token)
            except TokenExpired:
                token, refresh_token = get_refresh_token(params, refresh_token)
                candidate = fetch_and_add_resume(candidate, token)
            yield candidate


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    DR_PROFILES_URL = parameters.environment_url.rstrip("/")
    url = DIGITAL_RECRUITERS_WRITE_PROFILES_ENDPOINT.format(
        url_environnement=DR_PROFILES_URL, token=parameters.token
    )

    failed_profiles = []

    for profile in profiles:
        profile["reference"] = parameters.job_reference
        profile["ApplicationMessage"] = dict(
            message=parameters.message  # Candidate Message
        )
        response = requests.post(url, json=profile)

        if response.status_code == 201:
            adapter.info("Candidate profile pushed successfully.")
        elif response.status_code == 202:
            adapter.warning(
                "Candidate profile pushed, but some information is missing."
            )
        else:
            adapter.error(
                "Failed to push candidate profile. Status code: %s, Response: %s",
                response.status_code,
                response.text,
            )
            failed_profiles.append(profile)

    return failed_profiles


# Define the Warehouse for Digital Recruiters jobs and profiles
DigitalRecruitersJobWarehouse = Warehouse(
    name="DigitalRecruiters Jobs",
    data_schema=DigitalRecruitersJob,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read_jobs,
        endpoints=[
            ActionEndpoints(
                name="Read Jobs",
                description="Read jobs from Digital Recruiters",
                url=DIGITAL_RECRUITERS_JOBS_ENDPOINT,
            )
        ],
    ),
)

# Define the Warehouse for Digital Recruiters jobs and profiles
DigitalRecruitersReadProfilesWarehouse = Warehouse(
    name="DigitalRecruiters Read Profils",
    data_schema=DigitalRecruitersReadProfile,
    data_type=DataType.profile,
    read=WarehouseReadAction(
        parameters=ReadProfileParameters,
        function=read_profiles,
        endpoints=[
            ActionEndpoints(
                name="Read Profiles",
                description="Read profiles from Digital Recruiters",
                url=DIGITAL_RECRUITERS_READ_PROFILES_ENDPOINT,
            )
        ],
    ),
)
DigitalRecruitersWriteProfileWarehouse = Warehouse(
    name="DigitalRecruiters Write Profile",
    data_schema=DigitalRecruitersWriteProfile,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters,
        function=write,
        endpoints=[
            ActionEndpoints(
                name="Write Profile",
                description="Write profile to Digital Recruiters",
                url=DIGITAL_RECRUITERS_WRITE_PROFILES_ENDPOINT,
            )
        ],
    ),
)



import typing as t
from abc import ABC, abstractmethod

from pydantic import BaseModel


class BackendStore(ABC):
    @staticmethod
    @abstractmethod
    def NAME() -> str:
        raise NotImplementedError  # pragma: no cover

    @property
    def name(
        self,
    ) -> str:
        return self.NAME()

    @abstractmethod
    def save(self, key: str, data: BaseModel) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def load(self, key: str, parse_as: t.Type[BaseModel]) -> t.Optional[BaseModel]:
        raise NotImplementedError  # pragma: no cover



from typing import Optional

from pydantic import BaseModel


class CeridianDayforceJobModel(BaseModel):
    Title: str
    Description: Optional[str]
    ClientSiteName: Optional[str]
    ClientSiteXRefCode: Optional[str]
    CompanyName: Optional[str]
    ParentCompanyName: Optional[str]
    JobDetailsUrl: str
    ApplyUrl: Optional[str]
    AddressLine1: Optional[str]
    City: Optional[str]
    State: Optional[str]
    Country: Optional[str]
    PostalCode: Optional[str]
    DatePosted: Optional[str]
    LastUpdated: Optional[str]
    ReferenceNumber: int
    ParentRequisitionCode: Optional[int]
    IsVirtualLocation: Optional[bool]



from typing import Any, List, Optional

from pydantic import BaseModel, Field


class TeamtailorJobAttribute(BaseModel):
    title: str
    pitch: Optional[str]
    body: str = Field(..., description="job description")
    created_at: str = Field(..., alias="created-at")
    updated_at: Optional[str] = Field(..., alias="updated-at")
    status: Optional[str]
    tags: List[str]
    remote_status: Optional[str] = Field(..., alias="remote-status")
    currency: Optional[str]
    salary_time_unit: Optional[str] = Field(None, alias="salary-time-unit")
    min_salary: Optional[int] = Field(..., alias="min-salary")
    max_salary: Optional[int] = Field(..., alias="max-salary")
    employment_type: Optional[str] = Field(None, alias="employment-type")
    employment_level: Optional[str] = Field(None, alias="employment-level")
    internal: Optional[bool]
    start_date: Optional[str] = Field(None, alias="start-date")
    end_date: Optional[str] = Field(None, alias="end-date")


class TeamtailorJob(BaseModel):
    attributes: TeamtailorJobAttribute
    links: Optional[dict]
    id: Any


class TeamtailorCandidateAttribute(BaseModel):
    first_name: str = Field(..., alias="first-name")
    last_name: str = Field(..., alias="last-name")
    email: str
    phone: Optional[str]
    pitch: Optional[str] = Field(None, description="summary")
    resume: Optional[str] = Field(None, description="candidate profile resume")
    sourced: Optional[bool]
    tags: Optional[List]



import logging
import os
import typing as t
from io import BytesIO

import boto3
import botocore
from pydantic import BaseModel, create_model

from hrflow_connectors.core.backend.common import BackendStore

logger = logging.getLogger(__name__)


class S3Store(BackendStore):
    BUCKET_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_S3_BUCKET"
    PREFIX_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_S3_PREFIX"
    AWS_REGION_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_S3_AWS_REGION"
    AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_S3_AWS_ACCESS_KEY_ID"
    AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE = (
        "HRFLOW_CONNECTORS_S3_AWS_SECRET_ACCESS_KEY"
    )
    S3_FOLDER_NAME = "hrflow_connectors/store"

    def __init__(
        self,
    ) -> None:
        for environment_variable in [
            S3Store.BUCKET_ENVIRONMENT_VARIABLE,
            S3Store.AWS_REGION_ENVIRONMENT_VARIABLE,
        ]:
            if os.environ.get(environment_variable) is None:
                raise Exception(
                    "Missing environment variable {} in order to setup S3 store".format(
                        environment_variable
                    )
                )
        if os.environ.get(
            S3Store.AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE
        ) and os.environ.get(S3Store.AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE):
            logger.info("Configuring S3 Backend with explicit credentials")
            self.client = boto3.client(
                "s3",
                region_name=os.environ.get(S3Store.AWS_REGION_ENVIRONMENT_VARIABLE),
                aws_access_key_id=os.environ.get(
                    S3Store.AWS_ACCESS_KEY_ID_ENVIRONMENT_VARIABLE
                ),
                aws_secret_access_key=os.environ.get(
                    S3Store.AWS_SECRET_ACCESS_KEY_ENVIRONMENT_VARIABLE
                ),
            )
        else:
            logger.info("Configuring S3 Backend with implicit credentials")
            self.client = boto3.client(
                "s3",
                region_name=os.environ.get(S3Store.AWS_REGION_ENVIRONMENT_VARIABLE),
            )

        self.bucket = os.environ.get(S3Store.BUCKET_ENVIRONMENT_VARIABLE)
        self.key_prefix = S3Store.S3_FOLDER_NAME
        prefix = os.environ.get(S3Store.PREFIX_ENVIRONMENT_VARIABLE)
        if prefix is not None:
            self.key_prefix = "{}/{}".format(
                prefix.strip("/"), S3Store.S3_FOLDER_NAME.strip("/")
            )

        self.__check_store()

    @staticmethod
    def NAME() -> str:
        return "s3"

    def __check_store(self):
        root_model = create_model(
            "S3StoreRoot", root="HrFlow Connectors", store=S3Store.NAME()
        )
        root_key = "__root"
        root = root_model()
        try:
            self.save(root_key, root)
        except Exception as e:
            raise Exception("Failed to check writing to S3 with error={}".format(e))

        try:
            loaded = self.load(key=root_key, parse_as=root_model)
            assert loaded == root, "Loaded data should match original data"
        except Exception as e:
            raise Exception("Failed to check reading from S3 with error={}".format(e))
        logger.info("S3 Backend properly configured")

    def s3_key(self, key: str) -> str:
        return "{}/{}.json".format(self.key_prefix.strip("/"), key)

    def save(self, key: str, data: BaseModel) -> None:
        with BytesIO(initial_bytes=data.json().encode()) as raw:
            self.client.upload_fileobj(raw, self.bucket, self.s3_key(key))
        return None

    def load(self, key: str, parse_as: t.Type[BaseModel]) -> t.Optional[BaseModel]:
        with BytesIO() as raw:
            try:
                self.client.download_fileobj(self.bucket, self.s3_key(key), raw)
            except botocore.exceptions.ClientError as e:
                if int(e.response["Error"]["Code"]) == 404:
                    return None
                raise e
            return parse_as.parse_raw(raw.getvalue().decode())



import typing as t

from hrflow_connectors.connectors.hrflow.warehouse import HrFlowProfileParsingWarehouse
from hrflow_connectors.connectors.jobology.warehouse import JobologyProfilesWarehouse
from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)


def rename_profile_fields(jobology_profile: t.Dict) -> t.Dict:
    return {
        "job-key": jobology_profile["jobkey"],
        "first_name": jobology_profile.get("firstName"),
        "last_name": jobology_profile.get("lastName"),
        "phone": jobology_profile.get("phone"),
        "email": jobology_profile.get("email"),
        "coverText": jobology_profile.get("coverText"),
        "profile-country": jobology_profile.get("profilecountry"),
        "profile-regions": jobology_profile.get("profileregions"),
        "profile-domains": jobology_profile.get("profiledomains"),
        "job-lien_annonce_site_carriere": jobology_profile.get(
            "joblien_annonce_site_carriere"
        ),
        "statistic-source": jobology_profile.get("statisticsource"),
        "statistic-jbsource": jobology_profile.get("statisticjbsource"),
    }


def add_tags(profile_tags: t.Dict) -> t.List[t.Dict]:
    return [dict(name=key, value=value) for key, value in profile_tags.items() if value]


def format_jobology_profile(jobology_profile: t.List) -> t.Dict:
    profile_tags = rename_profile_fields(jobology_profile)
    tags = add_tags(profile_tags)
    resume_dict = dict(
        raw=jobology_profile["cv"],
        content_type=jobology_profile["content_type"],
    )
    return dict(
        reference=jobology_profile["email"],
        resume=resume_dict,
        tags=tags,
        metadatas=[],
        created_at=None,
    )


def event_parser(event: t.Dict) -> t.Dict:
    return dict(profile=event)


DESCRIPTION = (
    "La mission de jobology est de faciliter le processus de recrutement pour les"
    " entreprises "
)
Jobology = Connector(
    name="Jobology",
    type=ConnectorType.JobBoard,
    description=DESCRIPTION,
    url="https://www.jobology.fr/",
    actions=[
        ConnectorAction(
            name=ActionName.catch_profile,
            trigger_type=WorkflowType.catch,
            description="Imports candidates, in synchronization with jobology",
            parameters=BaseActionParameters.with_defaults(
                "TriggerViewActionParameters",
                format=format_jobology_profile,
                event_parser=event_parser,
            ),
            origin=JobologyProfilesWarehouse,
            target=HrFlowProfileParsingWarehouse,
            action_type=ActionType.inbound,
        )
    ],
)



import typing as t
from enum import Enum, IntEnum
from logging import LoggerAdapter

import requests
from pydantic import Field, validator

from hrflow_connectors.connectors.poleemploi.schemas import (
    ExperienceRequirement,
    OfferOriginTag,
    PoleEmploiJobOffer,
    validate_date,
)
from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
)
from hrflow_connectors.core.warehouse import ActionEndpoints

POLEEMPLOI_JOBS_SEARCH_ENDPOINT = (
    "https://api.emploi-store.fr/partenaire/offresdemploi/v2/offres/search"
)

POLEEMPLOI_REFERENCES_ENDPOINT = (
    "https://api.emploi-store.fr/partenaire/offresdemploi/v2/referentiel/"
)
SEARCH_JOBS_ENDPOINT = ActionEndpoints(
    name="Get all jobs",
    description=(
        "Endpoint to search offers based on selection criteria. The list of returned"
        " offers is paginated.The range of results is limited to 150.The request method"
        " is `GET`"
    ),
    url=(
        "https://pole-emploi.io/data/api/offres-emploi?tabgroup-api="
        "documentation&doc-section=api-doc-section-rechercher-par-crit%C3%A8res"
    ),
)
TOKEN_GENERATOR_URL = (
    "https://entreprise.pole-emploi.fr/connexion/oauth2/access_token?realm=/partenaire"
)

GRANT_TYPE = "client_credentials"
TOKEN_SCOPE = "api_offresdemploiv2 o2dsoffre"

MOTS_CLES_DESCRIPTION = """Search by keyword

Each keyword (or expression) is at least 2 characters long and must
be separated by a comma.
The search on several keywords is processed via the logical operator "AND".
The keyword search can be used to search on :

- The title of the offer (title field in the search return)
- The ROME code (romeCode field in the search return)
- The ROME label (field romeLibelle in return for the search)
- The competences label (field competences.libelle in return of the search)
- The wording of the training fields (field formations.domaineLibelle in
return of the research)
- The wording of the permits (field permits.label in return of the search)
- The language label (field languages.label in return of the search)
- The offer description if found in the offer title and/or the ROME label
(description field in the search return)

Allowed characters: [aA-zZ]+[0-9]+[space]+[@#$%^&+./-""]"""

MODESELECTIONPARTENAIRES_DESCRIPTION = """Selection mode of the partner offers.

This filter works with the partner criterion and is dependent on the originOffer
criterion. Possible values with the results obtained according to the two other filters:

- INCLUS(INCLUDED)
originOffer empty : Returns the PE offers and the Partners listed in the Partners
criterion
originOffer at 2 : Only the offers of the Partners listed in the Partners
criterion
- EXCLU(EXCLUDED)
originOffer empty : Return the offers of PE and Partners not listed in the Partners
criterion
originOffer at 2 : Only the offers of the Partners not listed in the Partners
criterion
Note: In all cases, if originOffer = 1, then only the Pole Emploi offers
will be returned"""


class Experience(str, Enum):
    LESS_THAN_ONE_YEAR = "1"
    ONE_TO_THREE_YEARS = "2"
    MORE_THAN_THREE_YEARS = "3"


class Qualification(IntEnum):
    NON_EXECUTIVE = 0
    EXECUTIVE = 9


class SalaryPeriod(str, Enum):
    MONTHLY = "M"
    ANNUALLY = "A"
    HOURLY = "H"
    FEE_FOR_SERVICE = "C"


class IndustryDomain(str, Enum):
    AGRICULTURE = "A"
    ARTS = "B"
    BANKING_INSURANCE = "C"
    REAL_ESTATE = "C15"
    RETAIL = "D"
    MEDIA_COMMUNICATION = "E"
    CONSTRUCTION = "F"
    HOSPITALITY_TOURISM = "G"
    INDUSTRY = "H"
    INSTALLATION_MAINTENANCE = "I"
    HEALTHCARE = "J"
    PERSONAL_COMMUNITY_SERVICES = "K"
    PERFORMING_ARTS = "L"
    SPORT = "L14"
    PURCHASING_ACCOUNTING_MANAGEMENT = "M"
    EXECUTIVE_MANAGEMENT = "M13"
    CONSULTING_RESEARCH = "M14"
    HUMAN_RESOURCES = "M15"
    SECRETARIAL_ASSISTANTSHIP = "M16"
    MARKETING_SALES_STRATEGY = "M17"
    COMPUTER_TELECOMMUNICATION = "M18"
    TRANSPORTATION_LOGISTICS = "N"


class PublishedSince(IntEnum):
    ONE_DAY = 1
    THREE_DAYS = 3
    SEVEN_DAYS = 7
    FOURTEEN_DAYS = 14
    THIRTY_ONE_DAYS = 31


class PartnerSelectionMode(str, Enum):
    INCLUDE = "INCLUS"
    EXCLUDE = "EXCLU"


class WeeklyDuration(str, Enum):
    NOT_SPECIFIED = "0"
    FULL_TIME = "1"
    PART_TIME = "2"


class ReadJobsParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client ID used to access Pole Emploi API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_secret: str = Field(
        ...,
        description="Client Secret used to access Pole Emploi API",
        repr=False,
        field_type=FieldType.Auth,
    )
    range: t.Optional[str] = Field(
        description="Pagination of data. The range of results is limited to 150.",
        field_type=FieldType.QueryParam,
    )
    sort: t.Optional[int] = Field(
        description="Sorting of data",
        field_type=FieldType.QueryParam,
    )
    domaine: t.Optional[str] = Field(
        description=(
            "Professional field code"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f"{POLEEMPLOI_REFERENCES_ENDPOINT}/domaines"
        ),
        field_type=FieldType.QueryParam,
    )
    codeROME: t.Optional[str] = Field(
        description=(
            "ROME code of the profession"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/metiers"
        ),
        field_type=FieldType.QueryParam,
    )
    theme: t.Optional[str] = Field(
        description=(
            "Theme of the profession"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/themes"
        ),
        field_type=FieldType.QueryParam,
    )
    appellation: t.Optional[str] = Field(
        description=(
            "Code of the appellation"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/appellations"
        ),
        field_type=FieldType.QueryParam,
    )
    secteurActivite: t.Optional[str] = Field(
        description=(
            "NAF codes for sectors of activity. It is possible to specify two NAF codes"
            " by separating them with a comma in the character string."
            "Example : 01,02"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/secteursActivites"
        ),
        field_type=FieldType.QueryParam,
    )
    experience: t.Optional[Experience] = Field(
        description=(
            "Level of experience required\nPossible values:\n1 -> Less than 1 year of"
            " experience\n2 -> From 1 to 3 years of experience\n3 -> More than 3 years"
            " of experience"
        ),
        field_type=FieldType.QueryParam,
    )
    typeContrat: t.Optional[str] = Field(
        description=(
            "Contract type code"
            "Example : CDI,CDD"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/typesContrats"
        ),
        field_type=FieldType.QueryParam,
    )
    natureContrat: t.Optional[str] = Field(
        description=(
            "Code of the nature of contract"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            " to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/naturesContrats"
        ),
        field_type=FieldType.QueryParam,
    )
    origineOffre: t.Optional[OfferOriginTag] = Field(
        description=(
            "Origin of the offer\nPossible values:\n1 -> Job center\n2 -> Partner"
        ),
        field_type=FieldType.QueryParam,
    )
    qualification: t.Optional[Qualification] = Field(
        description=(
            "Qualification Code\nPossible values:\n0 -> Non-executive\n9 -> Executive"
        ),
        field_type=FieldType.QueryParam,
    )
    tempsPlein: t.Optional[bool] = Field(
        description=(
            "Promote the use of the WeeklyDuration filter\nPossible values:\nfalse ->"
            " Part-time\ntrue -> Full time\nIf the parameter is not filled, then all"
            " the offers are returned"
        ),
        field_type=FieldType.QueryParam,
    )
    commune: t.Optional[str] = Field(
        description=(
            "INSEE code of the commune"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/communes"
        ),
        field_type=FieldType.QueryParam,
    )
    distance: t.Optional[int] = Field(
        description=(
            "Kilometric distance of the search radius\nDefault value: 10Note: to obtain"
            " only the offers of a specific commune, then you must fill in the"
            " parameter 'distance=0'."
        ),
        field_type=FieldType.QueryParam,
    )
    # distance=0 pour pour obtenir seulement les offres d'une commune spécifique
    departement: t.Optional[str] = Field(
        description=(
            "INSEE code of the department"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/departements"
        ),
        field_type=FieldType.QueryParam,
    )
    inclureLimitrophes: t.Optional[bool] = Field(
        description="Include bordering departments in the search",
        field_type=FieldType.QueryParam,
    )
    region: t.Optional[str] = Field(
        description=(
            "Code of the region of the offer"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/regions"
        ),
        field_type=FieldType.QueryParam,
    )
    paysContinent: t.Optional[str] = Field(
        description=(
            "Code of the country or continent of the offer"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/pays"
            f"AND {POLEEMPLOI_REFERENCES_ENDPOINT}/continents"
        ),
        field_type=FieldType.QueryParam,
    )
    niveauFormation: t.Optional[str] = Field(
        description=(
            "Level of education required"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/niveauxFormations"
        ),
        field_type=FieldType.QueryParam,
    )
    permis: t.Optional[str] = Field(
        description=(
            "Code of the requested license"
            "A GET request for the list of accepted choices from the Offres"
            " d'emploi API"
            "to this endpoint :"
            f" {POLEEMPLOI_REFERENCES_ENDPOINT}/permis"
        ),
        field_type=FieldType.QueryParam,
    )
    motsCles: t.Optional[str] = Field(
        description=MOTS_CLES_DESCRIPTION,
        field_type=FieldType.QueryParam,
    )
    salaireMin: t.Optional[float] = Field(
        description=(
            "Minimum wage, expressed in Euro.If this data is filled in, the code of"
            " the type of minimum wage is mandatory."
        ),
        field_type=FieldType.QueryParam,
    )  # If this field is set, then periodeSalaire is required.
    periodeSalaire: t.Optional[SalaryPeriod] = Field(
        description=(
            "Period for the calculation of the minimum wage.\nIf this data is filled"
            " in, the minimum wage is mandatory.\nPossible values:\nM -> Monthly\nA ->"
            " Annual\nH -> Hourly\nC -> Fee"
        ),
        field_type=FieldType.QueryParam,
    )  # If this field is set, then salaireMin is required.
    accesTravailleurHandicape: t.Optional[bool] = Field(
        description=(
            "Allows you to search for offers for which the employer is handi friendly"
        ),
        field_type=FieldType.QueryParam,
    )
    offresMRS: t.Optional[bool] = Field(
        description=(
            " Allows you to search for jobs that offer the simulation recruitment"
            " method"
        ),
        field_type=FieldType.QueryParam,
    )
    grandDomaine: t.Optional[IndustryDomain] = Field(
        description="Code of the major area of the offer",
        field_type=FieldType.QueryParam,
    )
    experienceExige: t.Optional[ExperienceRequirement] = Field(
        description="Filter offers by experience level.",
        field_type=FieldType.QueryParam,
    )
    publieeDepuis: t.Optional[PublishedSince] = Field(
        description=(
            "Maximum number of days since the publication of the offer\nPossible"
            " values: 1, 3, 7, 14, 31"
        ),
        field_type=FieldType.QueryParam,
    )
    minCreationDate: t.Optional[str] = Field(
        description=(
            "Minimum offer creation date.\nIf this data is filled in, the maximum offer"
            " creation date is mandatory.\nISO-8601 standard (YYYY-MM-DDTHH:MM:SSZ)"
        ),
        field_type=FieldType.QueryParam,
    )  # If this field is set, then minCreationDate is required.
    maxCreationDate: t.Optional[str] = Field(
        description=(
            "Maximum offer creation date.\nIf this data is filled in, the minimum offer"
            " creation date is mandatory.\nISO-8601 standard (YYYY-MM-DDTHH:MM:SSZ)"
        ),
        field_type=FieldType.QueryParam,
    )  # If this field is set, then maxCreationDate is required.

    partenaires: t.Optional[str] = Field(
        description=(
            " This filter allows you to enter your partner code in order to include or"
            " exclude your offers from the results according to the selectionmade in"
            " the PartnerSelection mode filter\nIt is possible to enter several codes"
            " (separator ','). "
        ),
        field_type=FieldType.QueryParam,
    )  # Il est possible de saisir plusieurs codes (séparateur ",").
    modeSelectionPartenaires: t.Optional[PartnerSelectionMode] = Field(
        description=MODESELECTIONPARTENAIRES_DESCRIPTION,
        field_type=FieldType.QueryParam,
    )  # If this field is set, then maxCreationDate is required.

    dureeHebdo: t.Optional[WeeklyDuration] = Field(
        description=(
            "Filtre les offres selon la durée hebdomadaire.\nValeurs possibles :\n0 ->"
            " Non précisé\n1 -> Temps plein\n2 -> Temps partiel"
        ),
        field_type=FieldType.QueryParam,
    )
    dureeHebdoMin: t.Optional[int] = Field(
        description=(
            "Minimum weekly duration of the offer\nThe value must be in HHMM format,"
            " for example : 8h => 800 ; 24h30 => 2430"
        ),
        field_type=FieldType.QueryParam,
    )  # format HHMM
    dureeHebdoMax: t.Optional[int] = Field(
        description=(
            "Maximum weekly duration of the offer\nThe value must be in HHMM format,"
            " for example: 8h => 800; 24h30 => 2430"
        ),
        field_type=FieldType.QueryParam,
    )  # format HHMM
    dureeContratMin: t.Optional[float] = Field(
        description=(
            "Minimum duration of the sought contract.\nThe search is done in months"
            " (ex: 0.5 for 15 days, 1.0 for 1 month,2.0 for 2 months).\nPositive"
            " decimal (Decimal separator: '.')"
        ),
        field_type=FieldType.QueryParam,
    )
    dureeContratMax: t.Optional[float] = Field(
        description=(
            "Maximum duration of the sought contract.\nThe search is made in months"
            " (ex: 0.5 for 15 days, 1.0 for 1 month,2.0 for 2 months).\nPositive"
            " decimal (Decimal separator: '.')"
        ),
        field_type=FieldType.QueryParam,
    )
    offresManqueCandidats: t.Optional[bool] = Field(
        description=(
            "Filters offers older than 15 days, with less than 4 applications (of which"
            " Pôle emploi is informed)\nfalse -> Offers not concerned\ntrue -> Offers"
            " with few candidates"
        ),
        field_type=FieldType.QueryParam,
    )
    entreprisesAdaptees: t.Optional[bool] = Field(
        description=(
            "Filter the offers where the adapted company allows a disabled worker to"
            " exercise a professional activity in  conditions adapted to his"
            " capacities\nfalse -> Offers not concerned\ntrue -> Offers from adapted"
            " companies"
        ),
        field_type=FieldType.QueryParam,
    )

    # validators
    _validate_minCreationDate = validator("minCreationDate", allow_reuse=True)(
        validate_date
    )
    _validate_maxCreationDate = validator("maxCreationDate", allow_reuse=True)(
        validate_date
    )


def get_poleemploi_auth_token(client_id: str, client_secret: str) -> str:
    response = requests.post(
        TOKEN_GENERATOR_URL,
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        data=dict(
            grant_type=GRANT_TYPE,
            scope=TOKEN_SCOPE,
            client_id=client_id,
            client_secret=client_secret,
        ),
    )
    if not response.ok:
        raise Exception(
            "Failed to get authentication token with error={}".format(response.text)
        )
    try:
        return response.json()["access_token"]
    except (KeyError, requests.exceptions.JSONDecodeError) as e:
        raise Exception(
            "Failed to get token from response with error={}".format(repr(e))
        )


def read(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    token = get_poleemploi_auth_token(
        client_id=parameters.client_id,
        client_secret=parameters.client_secret,
    )
    params = parameters.dict(exclude_none=True)
    del params["client_id"]
    del params["client_secret"]

    response = requests.get(
        POLEEMPLOI_JOBS_SEARCH_ENDPOINT,
        headers={"Authorization": "Bearer {}".format(token)},
        params=params,
    )
    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to pull jobs from Pole Emploi params={}"
            " status_code={} response={}".format(
                params, response.status_code, response.text
            )
        )
        raise Exception("Failed to pull jobs from Pole Emploi")
    jobs = response.json()["resultats"]
    for job in jobs:
        yield job


PoleEmploiJobWarehouse = Warehouse(
    name="Pole Emploi Jobs",
    data_schema=PoleEmploiJobOffer,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read,
        endpoints=[SEARCH_JOBS_ENDPOINT],
    ),
)



import datetime
import typing as t
from enum import Enum

from pydantic import BaseModel, validator


class Category(BaseModel):
    __CLASS__: str
    tag: str
    label: str


class Location(BaseModel):
    __CLASS__: str
    area: t.List[str]
    display_name: str


class ContractType(str, Enum):
    PERMANENT = "permanent"
    CONTRACT = "contract"


class Flag(str, Enum):
    YES = "1"
    NO = "0"


class ContractTime(str, Enum):
    FULL_TIME = "full_time"
    PART_TIME = "part_time"


class Company(BaseModel):
    __CLASS__: str
    display_name: str
    canonical_name: t.Optional[str]
    count: t.Optional[int]


class AdzunaJob(BaseModel):
    id: str
    created: str
    title: str
    description: str
    full_description: t.Optional[str]
    redirect_url: str
    latitude: t.Optional[float]
    longitude: t.Optional[float]
    category: Category
    location: Location
    salary_min: int
    salary_max: int
    salary_is_predicted: Flag
    company: Company
    contract_type: t.Optional[ContractType]
    contract_time: ContractTime

    @validator("created")
    def validate_date(value: t.Any) -> str:
        try:
            _ = datetime.date.fromisoformat(value)
            return value
        except ValueError:
            raise ValueError(
                "Invalid time format given {}, expected format ISO 8601 format"
                " ISO-8601 (YYYY-MM-DDTHH:MM:SSZ)".format(value)
            )



import typing as t

from hrflow_connectors.connectors.hrflow.warehouse import HrFlowProfileWarehouse
from hrflow_connectors.connectors.waalaxy.warehouse import WaalaxyProfilesWarehouse
from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)


def format_waalaxy_profile(waalaxy_profile: t.Dict) -> t.Dict:
    tags = [
        dict(name="prospectList", value=waalaxy_profile["prospectList"]),
        dict(name="messageSent", value=waalaxy_profile["messageSent"]),
        dict(name="messageReplied", value=waalaxy_profile["messageReplied"]),
        dict(name="emailSent", value=waalaxy_profile["emailSent"]),
        dict(name="emailReplied", value=waalaxy_profile["emailReplied"]),
    ]
    info = dict(
        first_name=waalaxy_profile["firstName"],
        last_name=waalaxy_profile["lastName"],
        full_name=" ".join([waalaxy_profile["firstName"], waalaxy_profile["lastName"]]),
        email=waalaxy_profile["email"],
        phone=waalaxy_profile["phoneNumbers"],
        location={"text": waalaxy_profile["location"], "lat": "", "lng": ""},
        urls=[
            dict(type="linkedin", url=waalaxy_profile["linkedinUrl"]),
            dict(type="company_linkedin", url=waalaxy_profile["company_linkedinUrl"]),
            dict(type="company_website", url=waalaxy_profile["company_website"]),
            dict(type="salesNavigator", url=waalaxy_profile["salesNavigatorUrl"]),
        ],
        summary=waalaxy_profile["occupation"],
    )

    return dict(
        reference=waalaxy_profile["_id"],
        tags=tags,
        info=info,
        text_language="en",
        experiences=[],
        educations=[],
    )


def event_parser(event: t.Dict) -> t.Dict:
    return dict(profile=event)


DESCRIPTION = "Waalaxy"
Waalaxy = Connector(
    name="Waalaxy",
    type=ConnectorType.Automation,
    description=DESCRIPTION,
    url="https://www.waalaxy.com/fr/",
    actions=[
        ConnectorAction(
            name=ActionName.catch_profile,
            trigger_type=WorkflowType.catch,
            description=(
                "Imports the visited profiles, in synchronization with the Waalaxy"
                " campaign (Visit + CRM Sync)"
            ),
            parameters=BaseActionParameters.with_defaults(
                "TriggerViewActionParameters",
                format=format_waalaxy_profile,
                event_parser=event_parser,
            ),
            origin=WaalaxyProfilesWarehouse,
            target=HrFlowProfileWarehouse,
            action_type=ActionType.inbound,
        )
    ],
)



# This code originally came from Legacy-master branch
# https://github.com/Riminder/hrflow-connectors/blob/legacy-master-ts-1663850377/src/hrflow_connectors/utils/adress_to_lat_long.py
# and was directly copied into the current branch without any changes.
import pkgutil
import re

import requests

# French Departements File
FRENCH_DEPARTMENTS_FILE = "../data/french_departement_geo_mapping.csv"
# French Cities File
FRENCH_CITIES_FILE = "../data/french_citycode_geo_mapping.csv"
# French Cities File
FRENCH_CITIES_NAME_FILE = "../data/french_cityname_geo_mapping.csv"


def accent_replacer(s: str) -> str:
    """
    Replace the accentuated characters by same character without accent.

    Args:
        s: the string we want to remove accentuated letters.

    Returns:
        The unaccentuated string.
    """
    s = s.replace("é", "e")
    s = s.replace("è", "e")
    s = s.replace("ê", "e")
    s = s.replace("ë", "e")
    s = s.replace("à", "a")
    s = s.replace("â", "a")
    s = s.replace("ä", "a")
    s = s.replace("ï", "i")
    s = s.replace("î", "i")
    s = s.replace("ù", "u")
    s = s.replace("û", "u")
    s = s.replace("ü", "u")
    s = s.replace("ç", "c")
    return s


def get_cities_code_lat_long_mapping() -> dict:
    """
    Get the dict of all postcode of french cities.

    Returns:
        exemple : ville_code = {"47380": ("44.525200649300004", "0.476034019359")}
    """
    ville_code = dict()
    # name used for the print.
    ville_code["name"] = "cities_codes"
    data = pkgutil.get_data(__name__, "../data/french_citycode_geo_mapping.csv")
    str_data = data.decode("utf-8")
    for line in str_data.split("\n"):
        row = line.split(",")
        code_str = str(row[0])
        # fill the uncompleted zip codes with zeros. (1 400 => 01 400)
        code = code_str.zfill(5)
        ville_code[code] = (row[1], row[2])

    return ville_code


def get_cities_names_lat_long_mapping() -> dict:
    """
    Get the dict of all names of french cities.

    Returns:
        exemple : ville_name = {"bergerac": ("44.854375187200006", "0.486529423457")}
    """
    ville_name = dict()
    # name used for the print.
    ville_name["name"] = "cities_name"
    data = pkgutil.get_data(__name__, "../data/french_cityname_geo_mapping.csv")
    str_data = data.decode("utf-8")
    for line in str_data.split("\n"):
        # .lower() for case sensitivity and accent_replacer(s) for accents.
        row = line.split(",")
        ville_name[accent_replacer(row[0].lower())] = (row[1], row[2])

    return ville_name


def get_departments_codes_lat_long_mapping() -> dict:
    """
    Get the dict of all postcode of french departments.

    Returns:
        exemple : departement = {"47": ("44.34453", "0.94193")}
    """
    departement = dict()
    # name used for the print.
    departement["name"] = "departments_codes"
    data = pkgutil.get_data(__name__, "../data/french_departement_geo_mapping.csv")
    str_data = data.decode("utf-8")
    for line in str_data.split("\n"):
        row = line.split(",")
        departement[row[0]] = (row[1], row[2])

    return departement


# Constant Dicts to avoid loading the files each time.
CITIES_CODES_DICT = get_cities_code_lat_long_mapping()
CITIES_NAMES_DICT = get_cities_names_lat_long_mapping()
DEPARTMENTS_CODES_DICT = get_departments_codes_lat_long_mapping()


def get_geolocation_data(
    location: str,
    cities_codes_dict: dict = None,
    cities_names_dict: dict = None,
    departments_codes_dict: dict = None,
    api_key=None,
):
    """
    Get the tuple latitude, longitude of a location with a 4 level of fall back.
        level 1: cities's postcode.
        level 2: cities's name.
        level 3: departments postcode.
        level 4: call here API.

    Args:
        location: the string of the location we want to get latitude and longitude.
        api_key: Here's api_key permitting the last fall back.
        cities_codes_dict:
        cities_names_dict:
        departments_codes_dict:

    Returns:
        latitude, longitude of the location.
    """
    cities_codes_dict = cities_codes_dict or CITIES_CODES_DICT
    cities_names_dict = cities_names_dict or CITIES_NAMES_DICT
    departments_codes_dict = departments_codes_dict or DEPARTMENTS_CODES_DICT

    fall_back_list = [cities_codes_dict, cities_names_dict, departments_codes_dict]
    for fall_back in fall_back_list:
        # get each word of the location.
        words_list = re.split("[,;._()/ ]+", accent_replacer(location))
        for word in words_list:
            value = fall_back.get(word.lower())
            if value:
                name = fall_back["name"]
                lat, long = value
                return name, lat, long

    if api_key is None:
        return None, None, None

    # If provided api_key, call here API to decode.
    URL = "https://geocode.search.hereapi.com/v1/geocode"
    PARAMS = {"apikey": api_key, "q": location}
    # sending get request and saving the response as response object
    r = requests.get(url=URL, params=PARAMS)
    data = r.json()
    if data.get("items"):
        # Acquiring the latitude and longitude from JSON
        latitude = data["items"][0]["position"]["lat"]
        longitude = data["items"][0]["position"]["lng"]
        return "here", latitude, longitude
    return None, None, None



import typing as t

from pydantic import BaseModel, Field


class Properties(BaseModel):
    email: str
    firstname: str
    lastname: str
    date_of_birth: t.Optional[str]
    phone: t.Optional[str]
    address: t.Optional[str]
    zip: t.Optional[str]
    city: t.Optional[str]
    state: t.Optional[str]
    country: t.Optional[str]
    jobtitle: t.Optional[str]
    company: t.Optional[str]
    annualrevenue: t.Optional[str]
    website: t.Optional[str]


class ContactObject(BaseModel):
    properties: Properties = Field(
        ...,
        description=(
            "Contact details are stored in contact properties. In addition to default"
            " properties, you can store custom data by creating custom contact"
            " properties. These can be managed through the CRM object properties"
            " endpoints."
        ),
    )



import json
import typing as t
from datetime import datetime, timezone
from logging import LoggerAdapter

from pydantic import Field
from simple_salesforce import Salesforce, SalesforceError
from simple_salesforce.bulk import SFBulkType

from hrflow_connectors.connectors.salesforce.schemas import (
    SalesforceHrFlowJob,
    SalesforceHrFlowProfile,
)
from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)

DEFAULT_LIMIT_PROFILES = 100
DEFAULT_LIMIT_JOBS = 1000
SOQL_MAX_RETURNED_ROWS = 2000

SELECT_PROFILES_SOQL = """
SELECT
    LastModifiedDate,
    Id__c,
    Dataset_Id__c,
    Date_Edition__c,
    Date_Reception__c,
    Text__c,
    Email__c,
    URLs__c,
    Name__c,
    Gender__c,
    Location_Text__c,
    Location_Lat__c,
    Location_Lng__c,
    Location_Gmaps__c,
    Phone__c,
    Timestamp__c,
    Summary__c,
    Hash_Id__c,
    Text_Language__c,
    Reference__c,
    Skills__c,
    Languages__c,
    Interests__c,
    Seniority__c,
    Location_Fields__c,
    Experiences_Duration__c,
    Educations_Duration__c,
    Archive__c,
    Picture__c,
    First_Name__c,
    Last_Name__c,
    Labels__c,
    Metadatas__c,
    Tags__c,
    Certifications__c,
    Courses__c,
    Tasks__c,
    Date_Birth__c,
    (
        SELECT
            FIELDS(CUSTOM)
        FROM HrFlow_Profile_Experiences__r
        LIMIT 200
    ),
    (
        SELECT
            FIELDS(CUSTOM)
        FROM HrFlow_Profile_Educations__r
        LIMIT 200
    ),
    (
        SELECT
            FIELDS(CUSTOM)
        FROM HrFlow_Profile_Attachments__r
        LIMIT 200
    )
FROM HrFlow_Profile__c
WHERE LastModifiedDate >= {last_modified_date}
ORDER BY LastModifiedDate, Id__c
{limit_placeholder}
"""

SELECT_JOBS_SOQL = """
SELECT
    LastModifiedDate,
    Id__c,
    Name__c,
    Slug__c,
    URL__c,
    Summary__c,
    Status__c,
    Date_Edition__c,
    Skills__c,
    Archive__c,
    Hash_Id__c,
    Location_Text__c,
    Location_Lat__c,
    Location_Lng__c,
    Tags__c,
    Board_Id__c,
    Sections__c,
    Reference__c,
    Languages__c,
    Ranges_Float__c,
    Ranges_Date__c,
    Location_Gmaps__c,
    Location_Fields__c,
    Metadatas__c,
    Certifications__c,
    Courses__c,
    Tasks__c,
    Picture__c,
    Culture__c,
    Responsibilities__c,
    Requirements__c,
    Benefits__c,
    Interviews__c
FROM HrFlow_Job__c
WHERE LastModifiedDate >= {last_modified_date}
ORDER BY LastModifiedDate, Id__c
{limit_placeholder}
"""


class SalesforceBaseParameters(ParametersModel):
    sf_username: str = Field(
        ...,
        description="username used to access Salesforce API",
        repr=False,
        field_type=FieldType.Auth,
    )
    sf_password: str = Field(
        ...,
        description="password used to access Salesforce API",
        repr=False,
        field_type=FieldType.Auth,
    )
    sf_security_token: str = Field(
        ...,
        description=(
            "Security Token to access Salesforce API.See below for instructions: How"
            " Can I Find My Security Token and Use It in Data Loader | Salesforce"
            " Platform "
            " https://www.youtube.com/watch?v=nYbfxeSGKFM&ab_channel=SalesforceSupport"
        ),
        repr=False,
        field_type=FieldType.Auth,
    )
    sf_organization_id: str = Field(
        ...,
        description=(
            "See below for instructions: "
            "How to find your organization id "
            " https://help.salesforce.com/s/articleView?id=000385215&type=1"
        ),
        repr=False,
        field_type=FieldType.Auth,
    )


class ReadProfilesParameters(SalesforceBaseParameters):
    last_modified_date: t.Optional[str] = Field(
        None,
        description="Last modified date",
        field_type=FieldType.QueryParam,
    )
    limit: int = Field(
        DEFAULT_LIMIT_PROFILES,
        description=(
            "Total number of items to pull from Salesforce."
            "By default limiting to {}".format(DEFAULT_LIMIT_PROFILES)
        ),
        field_type=FieldType.QueryParam,
    )


class ReadJobsParameters(SalesforceBaseParameters):
    last_modified_date: t.Optional[str] = Field(
        None,
        description="Last modified date",
        field_type=FieldType.QueryParam,
    )
    limit: int = Field(
        DEFAULT_LIMIT_JOBS,
        description=(
            "Total number of items to pull from Salesforce."
            "By default limiting to {}".format(DEFAULT_LIMIT_JOBS)
        ),
        field_type=FieldType.QueryParam,
    )


def generic_read_factory(
    soql_query: str,
) -> t.Callable[
    [LoggerAdapter, ParametersModel, t.Optional[ReadMode], t.Optional[str]],
    t.Iterable[t.Dict],
]:
    def _read_items(
        adapter: LoggerAdapter,
        parameters: t.Union[ReadProfilesParameters, ReadJobsParameters],
        read_mode: t.Optional[ReadMode] = None,
        read_from: t.Optional[str] = None,
    ) -> t.Iterable[t.Dict]:
        sf = Salesforce(
            username=parameters.sf_username,
            password=parameters.sf_password,
            security_token=parameters.sf_security_token,
            organizationId=parameters.sf_organization_id,
        )
        if read_mode is ReadMode.sync:
            if parameters.last_modified_date is None:
                raise Exception("last_modified_date cannot be None in ReadMode.sync")
            last_modified_date = parameters.last_modified_date
        else:
            if parameters.last_modified_date is not None:
                raise Exception(
                    "last_modified_date should not be supplied in ReadMode.incremental"
                )
            if read_from:
                try:
                    read_from = json.loads(read_from)
                    last_modified_date = read_from["last_modified_date"]
                    last_id = read_from["hrflow_id"]
                except json.JSONDecodeError as e:
                    raise Exception(
                        f"Failed to JSON parse read_from={read_from} error={e}"
                    )
                except KeyError as e:
                    raise Exception(
                        "Failed to find expected key in"
                        f" read_from={read_from} error={repr(e)}"
                    )
            else:
                last_modified_date = (
                    datetime.now(tz=timezone.utc)
                    .replace(year=2020)
                    .replace(hour=0, minute=0, second=0, microsecond=0)
                    .isoformat()
                )
                last_id = 0

        remaining = parameters.limit
        query = soql_query.format(
            last_modified_date=last_modified_date,
            limit_placeholder=(
                "LIMIT {}".format(remaining + 5)
                if remaining + 5 < SOQL_MAX_RETURNED_ROWS
                else ""
            ),
        )
        for item in sf.query_all_iter(query):
            if remaining <= 0:
                break
            if (
                item["LastModifiedDate"] == last_modified_date
                and item["Id__c"] <= last_id
            ):
                # This is why limit_placeholder is set with remainig + 5
                # In order to account for first items that might be already pulled
                continue
            yield item
            remaining -= 1

    return _read_items


def delete_from_salesforce(data: t.List[t.Tuple[SFBulkType, t.List[str]]]) -> None:
    for sf_bulk_type, sf_ids in data:
        if len(sf_ids) > 0:
            sf_bulk_type.delete(data=[dict(Id=sf_id) for sf_id in sf_ids])


def write_profiles(
    adapter: LoggerAdapter,
    parameters: SalesforceBaseParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed = []
    sf = Salesforce(
        username=parameters.sf_username,
        password=parameters.sf_password,
        security_token=parameters.sf_security_token,
        organizationId=parameters.sf_organization_id,
    )
    for profile in profiles:
        experiences, experience_sf_ids = (
            profile.pop("HrFlow_Profile_Experiences__r", None),
            [],
        )
        educations, education_sf_ids = (
            profile.pop("HrFlow_Profile_Educations__r", None),
            [],
        )
        attachments, attachment_sf_ids = (
            profile.pop("HrFlow_Profile_Attachments__r", None),
            [],
        )

        try:
            sf_profile_id = sf.HrFlow_Profile__c.create(profile)["id"]
        except SalesforceError as e:
            adapter.error(
                "Failed to push profile(reference={}, id={}) to HrFlow_Profile__c with"
                " error={}".format(profile["Reference__c"], profile["Id__c"], e.content)
            )
            failed.append(profile)
            continue

        try:
            for related, sf_type, record_ids in [
                (experiences, sf.HrFlow_Experience__c, experience_sf_ids),
                (educations, sf.HrFlow_Education__c, education_sf_ids),
                (attachments, sf.HrFlow_Attachment__c, attachment_sf_ids),
            ]:
                records = related["records"] if related else []
                for i, record in enumerate(records):
                    record["Profile__c"] = sf_profile_id
                    record_ids.append(sf_type.create(record)["id"])
        except SalesforceError as e:
            adapter.error(
                "Failed to push related {} number={} for profile(reference={}, id={},"
                " sf_id={}) data={} with error={}. Cleaning profile and related from"
                " Salesforce and adding as failed item".format(
                    e.resource_name,
                    i,
                    profile["Reference__c"],
                    profile["Id__c"],
                    sf_profile_id,
                    record,
                    e.content,
                )
            )
            delete_from_salesforce(
                data=[
                    (sf.bulk.HrFlow_Profile__c, [sf_profile_id]),
                    (sf.bulk.HrFlow_Experience__c, experience_sf_ids),
                    (sf.bulk.HrFlow_Education__c, education_sf_ids),
                    (sf.bulk.HrFlow_Attachment__c, attachment_sf_ids),
                ]
            )
            failed.append(profile)
            continue

    return failed


def item_to_read_from(item: t.Dict) -> str:
    return json.dumps(
        dict(
            last_modified_date=item["LastModifiedDate"],
            hrflow_id=item["Id__c"],
        )
    )


SalesforceProfileWarehouse = Warehouse(
    name="Salesforce Profiles",
    data_schema=SalesforceHrFlowProfile,
    data_type=DataType.profile,
    read=WarehouseReadAction(
        parameters=ReadProfilesParameters,
        function=generic_read_factory(soql_query=SELECT_PROFILES_SOQL),
        supports_incremental=True,
        item_to_read_from=item_to_read_from,
    ),
    write=WarehouseWriteAction(
        parameters=SalesforceBaseParameters,
        function=write_profiles,
    ),
)

SalesforceJobWarehouse = Warehouse(
    name="Salesforce Jobs",
    data_schema=SalesforceHrFlowJob,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=generic_read_factory(soql_query=SELECT_JOBS_SOQL),
        supports_incremental=True,
        item_to_read_from=item_to_read_from,
    ),
)



# Code from PR: https://github.com/Riminder/hrflow-connectors/pull/66
# Author: Yasser BELMEHDI (yass1337)
import enum
import json
import re
import typing as t
from logging import LoggerAdapter

import requests
from pydantic import Field

from hrflow_connectors.connectors.taleez.schemas import Candidate, Job
from hrflow_connectors.core import (
    ActionEndpoints,
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)

POST_CANDIDATE_ENDPOINT = "https://api.taleez.com/0/candidates"
TALEEZ_JOBS_ENDPOINT = "https://api.taleez.com/0/jobs"
TALEEZ_JOBS_ENDPOINT_LIMIT = 100
ACCEPT = "application/json;charset=UTF-8"
CONTENT_TYPE = "application/json"
GET_ALL_JOBS_ENDPOINT = ActionEndpoints(
    name="Get all jobs",
    description=(
        "Endpoint to retrieve all jobs."
        " and get the list of all jobs with their ids, the request method"
        " is `GET`"
    ),
    url="https://api.taleez.com/0/jobs",
)


class JobStatus(str, enum.Enum):
    published = "PUBLISHED"


class ReadJobsParameters(ParametersModel):
    x_taleez_api_secret: str = Field(
        ...,
        description="X-taleez-api-secret used to access Taleez API",
        repr=False,
        field_type=FieldType.Auth,
    )
    with_details: bool = Field(..., description="xxx", field_type=FieldType.QueryParam)
    job_status: JobStatus = Field(
        None,
        description="Posting status of a job. One of {}".format(
            [e.value for e in JobStatus]
        ),
        field_type=FieldType.QueryParam,
    )


def read(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    params = dict(withDetails=parameters.with_details, status=parameters.job_status)

    response = requests.get(
        TALEEZ_JOBS_ENDPOINT,
        headers={"X-taleez-api-secret": parameters.x_taleez_api_secret},
        params=params,
    )

    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to pull jobs from Taleez params={}"
            " status_code={} response={}".format(
                params, response.status_code, response.text
            )
        )
        raise Exception("Failed to pull jobs from Taleez")

    response = response.json()
    jobs = response["list"]

    for job in jobs:
        yield job


class WriteProfilesParameters(ParametersModel):
    accept: str = Field(
        ACCEPT, const=True, field_type=FieldType.QueryParam
    )  # TODO: Include Headers in FieldType
    x_taleez_api_secret: str = Field(
        ...,
        description="Client Secret id used to access Taleez API",
        repr=False,
        field_type=FieldType.Auth,
    )
    content_type: str = Field(
        ..., description="Content type", repr=True, field_type=FieldType.QueryParam
    )


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    data: t.Iterable[t.Dict],
) -> t.Iterable[t.Dict]:
    data = list(data)
    adapter.info("Pushing {} candidates to Taleez API".format(len(data)))
    failed_profiles = []
    for profile in data:
        headers = {
            "accept": "{}".format(parameters.accept),
            "X-taleez-api-secret": "{}".format(parameters.x_taleez_api_secret),
            "Content-Type": "{}".format(parameters.content_type),
        }

        response = requests.post(
            POST_CANDIDATE_ENDPOINT,
            headers=headers,
            data=json.dumps(profile["candidate"]),
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to create candidate status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
        id = int(re.search("[0-9]{7,8}", str(response.content)).group(0))
        binary_cv = requests.get(profile["CV"]).content
        response = requests.post(
            "{}/{}/documents?cv=true".format(POST_CANDIDATE_ENDPOINT, id),
            headers={
                "accept": "{}".format(parameters.accept),
                "X-taleez-api-secret": "{}".format(parameters.x_taleez_api_secret),
            },
            files=[
                (
                    "file",
                    (
                        "CV_{}_{}".format(
                            profile["candidate"]["firstName"],
                            profile["candidate"]["lastName"],
                        ),
                        binary_cv,
                        "application/pdf",
                    ),
                )
            ],
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to add document status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
        if profile["properties"]:
            response = requests.post(
                "{}/{}/properties".format(POST_CANDIDATE_ENDPOINT, id),
                data=json.dumps(profile["properties"]),
                headers={
                    "accept": "{}".format(parameters.accept),
                    "X-taleez-api-secret": "{}".format(parameters.x_taleez_api_secret),
                    "Content-Type": "{}".format(parameters.content_type),
                },
            )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update property status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)

    return failed_profiles


TaleezProfilesWarehouse = Warehouse(
    name="Taleez Profiles Warehouse",
    data_schema=Candidate,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters,
        function=write,
    ),
)
TaleezJobWarehouse = Warehouse(
    name="Taleez Jobs Warehouse",
    data_schema=Job,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read,
        endpoints=[GET_ALL_JOBS_ENDPOINT],
    ),
)



import typing as t
from logging import LoggerAdapter

import requests
from pydantic import Field

from hrflow_connectors.connectors.bullhorn.schemas import BullhornJob, BullhornProfile
from hrflow_connectors.connectors.bullhorn.utils.authentication import auth
from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)
from hrflow_connectors.core.warehouse import ReadMode


class WriteProfilesParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    client_secret: str = Field(
        ...,
        description="Client secret identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    password: str = Field(
        ...,
        description="Passoword for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )

    username: str = Field(
        ...,
        description="Username for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )


class ReadJobsParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    client_secret: str = Field(
        ...,
        description="Client secret identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    password: str = Field(
        ...,
        description="Passoword for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )

    username: str = Field(
        ...,
        description="Username for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )


class ReadProfileParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    client_secret: str = Field(
        ...,
        description="Client secret identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    password: str = Field(
        ...,
        description="Passoword for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )

    username: str = Field(
        ...,
        description="Username for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    adapter.info("Pushing {} profiles".format(len(profiles)))
    failed_profiles = []
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )

    for profile in profiles:
        # Split profile in four parts: Body, Education, Experience and Attachements
        profile_body_dict = profile
        create_profile_body = profile_body_dict["create_profile_body"]
        enrich_profile_education = profile_body_dict["enrich_profile_education"]
        enrich_profile_experience = profile_body_dict["enrich_profile_experience"]
        enrich_profile_attachment = profile_body_dict["enrich_profile_attachment"]

        rest_url = authentication["restUrl"]
        params = {"BhRestToken": authentication["BhRestToken"]}

        candidate_url = rest_url + "entity/Candidate"
        response = requests.put(
            url=candidate_url, json=create_profile_body, params=params
        )

        # Unable to push profile
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to push profile from to Bullhorn"
                " status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            failed_profiles.append(profile)
            continue

        candidate_id = response.json()
        candidate_id = str(candidate_id["changedEntityId"])

        # Enrich profile education
        for education in enrich_profile_education:
            education["candidate"]["id"] = candidate_id
            education_url = rest_url + "entity/CandidateEducation"
            response = requests.put(url=education_url, json=education, params=params)

        # Enrich profile experience
        for experience in enrich_profile_experience:
            experience["candidate"]["id"] = candidate_id
            experience_url = rest_url + "entity/CandidateWorkHistory"
            response = requests.put(url=experience_url, json=experience, params=params)

        # Enrich profile attachements
        for attachment in enrich_profile_attachment:
            attachment_url = rest_url + "file/Candidate/" + str(candidate_id)
            response = requests.put(url=attachment_url, json=attachment, params=params)

    return failed_profiles


def read_jobs(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )
    start = 0

    while True:
        jobs_url = (
            authentication["restUrl"]
            + "search/JobOrder?query=(isOpen:true)&fields=*&BhRestToken="
            + authentication["BhRestToken"]
            + "&start="
            + str(start)
        )

        response = requests.get(url=jobs_url)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Bullhorn status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Bullhorn")
        response = response.json()

        start = response["start"] + response["count"]
        data = response["data"]

        for job in data:
            yield job

        if start >= response["total"]:
            break


def read_profiles_parsing(
    adapter: LoggerAdapter,
    parameters: ReadProfileParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )
    start = 0
    while True:
        profiles_url = (
            authentication["restUrl"] + "myCandidates?fields=*&start=" + str(start)
        )
        headers = {"BhRestToken": authentication["BhRestToken"]}

        response = requests.get(url=profiles_url, headers=headers)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from Bullhorn"
                " status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            raise Exception("Failed to pull profiles from Bullhorn")
        response = response.json()

        start = response["start"] + response["count"]
        data = response["data"]
        total = response["total"]

        for profile in data:
            profile["cvFile"] = None
            url_files = (
                authentication["restUrl"]
                + "entityFiles/Candidate/"
                + str(profile["id"])
            )
            headers = {"BhRestToken": authentication["BhRestToken"]}
            response = requests.get(url=url_files, headers=headers)
            response = response.json()

            last_cv = None
            curr_entity = None
            if len(response["EntityFiles"]) > 0:
                for entity_file in response["EntityFiles"]:
                    if entity_file["type"] == "Resume":
                        if not curr_entity:
                            curr_entity = entity_file
                            last_cv = entity_file["id"]
                        elif curr_entity["dateAdded"] < entity_file["dateAdded"]:
                            curr_entity = entity_file
                            last_cv = entity_file["id"]

            if last_cv is not None:
                url_cv = (
                    authentication["restUrl"]
                    + "/file/Candidate/"
                    + str(profile["id"])
                    + "/"
                    + str(last_cv)
                    + "/raw"
                )
                response = requests.get(url=url_cv, headers=headers)

                file = response.content
                with open("cv_file.pdf", "wb") as f:
                    f.write(file)

                with open("cv_file.pdf", "rb") as f:
                    profile_file = f.read()
                    profile["cvFile"] = profile_file

                yield profile

        if start >= total:
            break


def read_profiles(
    adapter: LoggerAdapter,
    parameters: ReadProfileParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )
    fields = """firstName,lastName,name,email,mobile,dateOfBirth,gender,address,dateAvailable,status,employeeType,activePlacements,
    skillSet,id,educations,workHistories"""
    start = 0
    while True:
        profiles_url = (
            authentication["restUrl"]
            + "myCandidates?fields="
            + fields
            + "&start="
            + str(start)
        )
        headers = {"BhRestToken": authentication["BhRestToken"]}

        response = requests.get(url=profiles_url, headers=headers)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from Bullhorn"
                " status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            raise Exception("Failed to pull profiles from Bullhorn")
        response = response.json()

        start = response["start"] + response["count"]
        total = response["total"]
        data = response["data"]

        for profile in data:
            # Enrich profile education
            education_ids = []
            educations = []
            for ed in profile["educations"]["data"]:
                education_ids.append(ed["id"])
            for id in education_ids:
                education_url = (
                    authentication["restUrl"]
                    + "entity/CandidateEducation/"
                    + str(id)
                    + "?fields=*"
                )
                response = requests.get(url=education_url, headers=headers)
                response = response.json()
                educations.append(response["data"])

            # Enrich profile work history
            work_history_ids = []
            work_histories = []
            for work_history in profile["workHistories"]["data"]:
                work_history_ids.append(work_history["id"])
            for id in work_history_ids:
                work_history_url = (
                    authentication["restUrl"]
                    + "entity/CandidateWorkHistory/"
                    + str(id)
                    + "?fields=*"
                )
                response = requests.get(url=work_history_url, headers=headers)
                response = response.json()
                work_histories.append(response["data"])

            profile["educations"] = educations
            profile["workHistories"] = work_histories

            yield profile

        if start >= total:
            break


BullhornProfileWarehouse = Warehouse(
    name="Bullhorn Profiles",
    data_schema=BullhornProfile,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters, function=write, endpoints=[]
    ),
    read=WarehouseReadAction(
        parameters=ReadProfileParameters, function=read_profiles, endpoints=[]
    ),
)

BullhornProfileParsingWarehouse = Warehouse(
    name="Bullhorn Profiles",
    data_schema=BullhornProfile,
    data_type=DataType.profile,
    read=WarehouseReadAction(
        parameters=ReadProfileParameters, function=read_profiles_parsing, endpoints=[]
    ),
)

BullhornJobWarehouse = Warehouse(
    name="Bullhorn Jobs",
    data_schema=BullhornJob,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read_jobs,
        endpoints=[],
    ),
)



import logging
import os

ACTIVE_STORES = []
from hrflow_connectors.core.backend.localjson import LocalJsonStore  # noqa

ACTIVE_STORES.append(LocalJsonStore)
try:
    from hrflow_connectors.core.backend.s3 import S3Store

    ACTIVE_STORES.append(S3Store)  # pragma: nocover
except ModuleNotFoundError:  # pragma: nocover
    pass  # pragma: nocover

logger = logging.getLogger(__name__)
store = None
is_configured = False

ENABLE_STORE_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_STORE_ENABLED"
STORE_NAME_ENVIRONMENT_VARIABLE = "HRFLOW_CONNECTORS_STORE"
DEFAULT_STORE = LocalJsonStore.NAME()


NAME_TO_STORE = {store.NAME(): store for store in ACTIVE_STORES}


def configure_store():
    global store, is_configured

    enable_store = os.environ.get(ENABLE_STORE_ENVIRONMENT_VARIABLE, None)
    if not enable_store or enable_store in ["false", "False", "0"]:
        logger.info("No backend configured. Incremental mode is not possible.")
        store = None
        is_configured = False
        return

    store_name = os.environ.get(STORE_NAME_ENVIRONMENT_VARIABLE, DEFAULT_STORE)
    try:
        store_class = NAME_TO_STORE[store_name]
    except KeyError:
        raise Exception(
            "{}='{}' is not a valid store use one of {}".format(
                STORE_NAME_ENVIRONMENT_VARIABLE,
                store_name,
                list(NAME_TO_STORE.keys()),
            )
        )

    logger.info("Starting {} Backend configuration".format(store_name))
    store = store_class()
    is_configured = True



import typing as t

from hrflow_connectors.connectors.adzuna.warehouse import AdzunaJobWarehouse
from hrflow_connectors.connectors.hrflow.warehouse import HrFlowJobWarehouse
from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)


def get_job_location(adzuna_job: t.Dict) -> t.Dict:
    lat = adzuna_job.get("latitude")
    lat = float(lat) if lat is not None else lat

    lng = adzuna_job.get("longitude")
    lng = float(lng) if lng is not None else lng

    return dict(lat=lat, lng=lng, text=adzuna_job.get("location")["display_name"])


def get_tags(adzuna_job: t.Dict) -> t.List[t.Dict]:
    job = adzuna_job
    tags = []
    salaries_are_predicted = job.get("salary_is_predicted") == "1"

    t = lambda name, value: dict(name=name, value=value)
    tags = [
        t("salary_min", job.get("salary_min")),
        t("salary_max", job.get("salary_max")),
        t("salaries_are_predicted", salaries_are_predicted),
        t("category", job.get("category", {}).get("label")),
        t("company", job.get("company", {}).get("display_name")),
    ]
    return tags


def format_job(
    adzuna_job: t.Dict,
) -> t.Dict:
    job = dict(
        name=adzuna_job.get("title"),
        reference=str(adzuna_job.get("id")),
        created_at=adzuna_job.get("created"),
        location=get_job_location(adzuna_job),
        url=adzuna_job.get("redirect_url"),
        summary=adzuna_job.get("description"),
        sections=[],
        tags=get_tags(adzuna_job),
    )
    return job


Adzuna = Connector(
    name="Adzuna",
    type=ConnectorType.JobBoard,
    description="Retrieve Adzuna's job, property and car advertisement listings.",
    url="https://www.adzuna.fr/",
    actions=[
        ConnectorAction(
            name=ActionName.pull_job_list,
            trigger_type=WorkflowType.pull,
            description=(
                "Retrieves jobs via the ***Adzuna'*** API Search endpoint"
                "and send them to a ***Hrflow.ai Board***."
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadJobsActionParameters", format=format_job
            ),
            origin=AdzunaJobWarehouse,
            target=HrFlowJobWarehouse,
            action_type=ActionType.inbound,
        ),
    ],
)



import json
import typing as t

from hrflow_connectors.connectors.hrflow.schemas import HrFlowProfile
from hrflow_connectors.connectors.hrflow.warehouse import (
    HrFlowJobWarehouse,
    HrFlowProfileWarehouse,
)
from hrflow_connectors.connectors.salesforce.schemas import (
    SalesforceHrFlowJob,
    SalesforceHrFlowProfile,
)
from hrflow_connectors.connectors.salesforce.warehouse import (
    SalesforceJobWarehouse,
    SalesforceProfileWarehouse,
)
from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)


def format_into_hrflow_profile(data: SalesforceHrFlowProfile) -> t.Dict:
    info = dict(
        full_name=f"{data['Last_Name__c']} {data['First_Name__c']}",
        first_name=data["First_Name__c"],
        last_name=data["Last_Name__c"],
        email=data["Email__c"],
        phone=data["Phone__c"],
        date_birth=data["Date_Birth__c"],
        location=dict(
            text=data["Location_Text__c"],
            lat=data["Location_Lat__c"],
            lng=data["Location_Lng__c"],
        ),
        gender=data["Gender__c"],
    )
    experiences = []
    if data["HrFlow_Profile_Experiences__r"]:
        experiences = [
            dict(
                title=experience["Title__c"],
                location=dict(
                    text=experience["Location_Text__c"],
                    lat=experience["Location_Lat__c"],
                    lng=experience["Location_Lng__c"],
                ),
                company=experience["Company__c"],
                date_start=dict(iso8601=experience["Date_Begin__c"]),
                date_end=dict(iso8601=experience["Date_End__c"]),
                description=experience["Description__c"],
                skills=json.loads(experience["Skills__c"]),
                tasks=json.loads(experience["Tasks__c"]),
                certifications=json.loads(experience["Certifications__c"]),
            )
            for experience in data["HrFlow_Profile_Experiences__r"]["records"]
        ]
    educations = []
    if data["HrFlow_Profile_Educations__r"]:
        educations = [
            dict(
                title=education["Title__c"],
                location=dict(
                    text=education["Location_Text__c"],
                    lat=education["Location_Lat__c"],
                    lng=education["Location_Lng__c"],
                ),
                school=education["School__c"],
                date_start=dict(iso8601=education["Date_Begin__c"]),
                date_end=dict(iso8601=education["Date_End__c"]),
                description=education["Description__c"],
                skills=json.loads(education["Skills__c"]),
                tasks=json.loads(education["Tasks__c"]),
                certifications=json.loads(education["Certifications__c"]),
                courses=json.loads(education["Courses__c"]),
            )
            for education in data["HrFlow_Profile_Educations__r"]["records"]
        ]
    attachments = []
    if data["HrFlow_Profile_Attachments__r"]:
        attachments = [
            dict(
                text=attachment["Text__c"],
                type=attachment["Type__c"],
                alt=attachment["Alt__c"],
                file_size=attachment["File_Size__c"],
                file_name=attachment["File_Name__c"],
                original_file_name=attachment["Original_File_Name__c"],
                extension=attachment["Extension__c"],
                url=attachment["URL__c"],
            )
            for attachment in data["HrFlow_Profile_Attachments__r"]["records"]
        ]
    return dict(
        key=data["Hash_Id__c"],
        reference=data["Reference__c"],
        archived_at=data["Archive__c"],
        updated_at=data["Date_Edition__c"],
        created_at=data["Date_Reception__c"],
        info=info,
        text_language=data["Text_Language__c"],
        text=data["Text__c"],
        educations_duration=data["Experiences_Duration__c"],
        experiences=experiences,
        educations=educations,
        attachments=attachments,
        skills=json.loads(data["Skills__c"]),
        languages=json.loads(data["Languages__c"]),
        certifications=json.loads(data["Certifications__c"]),
        courses=json.loads(data["Courses__c"]),
        tasks=json.loads(data["Tasks__c"]),
        interests=json.loads(data["Interests__c"]),
        labels=json.loads(data["Labels__c"]),
        tags=json.loads(data["Tags__c"]),
        metadatas=json.loads(data["Metadatas__c"]),
    )


def format_into_salesforce_profile(data: HrFlowProfile) -> t.Dict:
    experiences = None
    if data["experiences"]:
        experiences = dict(
            done=True,
            totalSize=len(data["experiences"]),
            records=[
                dict(
                    Certifications__c=json.dumps(experience["certifications"]),
                    Company__c=experience["company"],
                    Courses__c=json.dumps(experience["courses"]),
                    Date_Begin__c=experience["date_start"],
                    Date_End__c=experience["date_end"],
                    Description__c=experience["description"],
                    Hash_Id__c=experience["key"],
                    Location_Fields__c=json.dumps(experience["location"]["fields"]),
                    Location_Lat__c=experience["location"]["lat"],
                    Location_Lng__c=experience["location"]["lng"],
                    Location_Text__c=experience["location"]["text"],
                    Location_Gmaps__c=experience["location"]["gmaps"],
                    Skills__c=json.dumps(experience["skills"]),
                    Tasks__c=json.dumps(experience["tasks"]),
                    Title__c=experience["title"],
                )
                for experience in data["experiences"]
            ],
        )
    educations = None
    if data["educations"]:
        educations = dict(
            done=True,
            totalSize=len(data["educations"]),
            records=[
                dict(
                    Certifications__c=json.dumps(education["certifications"]),
                    School__c=education["school"],
                    Courses__c=json.dumps(education["courses"]),
                    Date_Begin__c=education["date_start"],
                    Date_End__c=education["date_end"],
                    Description__c=education["description"],
                    Hash_Id__c=education["key"],
                    Location_Fields__c=json.dumps(education["location"]["fields"]),
                    Location_Lat__c=education["location"]["lat"],
                    Location_Lng__c=education["location"]["lng"],
                    Location_Text__c=education["location"]["text"],
                    Location_Gmaps__c=education["location"]["gmaps"],
                    Skills__c=json.dumps(education["skills"]),
                    Tasks__c=json.dumps(education["tasks"]),
                    Title__c=education["title"],
                )
                for education in data["educations"]
            ],
        )
    attachments = None
    if data["attachments"]:
        attachments = dict(
            done=True,
            totalSize=len(data["attachments"]),
            records=[
                dict(
                    Alt__c=attachment["alt"],
                    Date_Edition__c=attachment["updated_at"],
                    Extension__c=attachment["extension"],
                    File_Name__c=attachment["file_name"],
                    File_Size__c=attachment["file_size"],
                    Original_File_Name__c=attachment["original_file_name"],
                    Timestamp__c=attachment["created_at"],
                    Type__c=attachment["type"],
                    URL__c=attachment["public_url"],
                )
                for attachment in data["attachments"]
            ],
        )
    return dict(
        Id__c=data["id"],
        Hash_Id__c=data["key"],
        Reference__c=data["reference"],
        Archive__c=data["archived_at"],
        Date_Edition__c=data["updated_at"],
        Date_Reception__c=data["created_at"],
        First_Name__c=data["info"]["first_name"],
        Last_Name__c=data["info"]["last_name"],
        Email__c=data["info"]["email"],
        Phone__c=data["info"]["phone"],
        Date_Birth__c=data["info"]["date_birth"],
        Location_Fields__c=json.dumps(data["info"]["location"]["fields"]),
        Location_Lat__c=data["info"]["location"]["lat"],
        Location_Lng__c=data["info"]["location"]["lng"],
        Location_Text__c=data["info"]["location"]["text"],
        Location_Gmaps__c=data["info"]["location"]["gmaps"],
        URLs__c=json.dumps(data["info"]["urls"]),
        Picture__c=data["info"]["picture"],
        Gender__c=data["info"]["gender"],
        Summary__c=data["info"]["summary"],
        Text_Language__c=data["text_language"],
        Text__c=data["text"],
        Experiences_Duration__c=data["experiences_duration"],
        Educations_Duration__c=data["educations_duration"],
        HrFlow_Profile_Experiences__r=experiences,
        HrFlow_Profile_Educations__r=educations,
        HrFlow_Profile_Attachments__r=attachments,
        Skills__c=json.dumps(data["skills"]),
        Languages__c=json.dumps(data["languages"]),
        Certifications__c=json.dumps(data["certifications"]),
        Courses__c=json.dumps(data["courses"]),
        Tasks__c=json.dumps(data["tasks"]),
        Interests__c=json.dumps(data["interests"]),
        Labels__c=json.dumps(data["labels"]),
        Tags__c=json.dumps(data["tags"]),
        Metadatas__c=json.dumps(data["metadatas"]),
    )


def format_job(data: SalesforceHrFlowJob) -> t.Dict:
    return dict(
        archived_at=data["Archive__c"],
        archive=data["Archive__c"],
        name=data["Name__c"],
        reference=data["Reference__c"],
        url=data["URL__c"],
        picture=data["Picture__c"],
        summary=data["Summary__c"],
        location=dict(
            text=data["Location_Text__c"],
            lat=data["Location_Lat__c"],
            lng=data["Location_Lng__c"],
        ),
        culture=data["Culture__c"],
        responsibilities=data["Responsibilities__c"],
        requirements=data["Requirements__c"],
        benefits=data["Benefits__c"],
        interviews=data["Interviews__c"],
        sections=json.loads(data["Sections__c"]),
        skills=json.loads(data["Skills__c"]),
        languages=json.loads(data["Languages__c"]),
        tags=json.loads(data["Tags__c"]),
        ranges_date=json.loads(data["Ranges_Date__c"]),
        ranges_float=json.loads(data["Ranges_Float__c"]),
        metadatas=json.loads(data["Metadatas__c"]),
    )


DESCRIPTION = (
    "Salesforce is the customer company. We make cloud-based software "
    "designed to help businesses connect to their customers in a whole "
    "new way, so they can find more prospects, close more deals, and wow"
    " customers with amazing service."
)

Salesforce = Connector(
    name="Salesforce",
    type=ConnectorType.CRM,
    description=DESCRIPTION,
    url="https://www.salesforce.com",
    actions=[
        ConnectorAction(
            name=ActionName.pull_profile_list,
            trigger_type=WorkflowType.pull,
            description=(
                "Retrieves profiles from Salesforce HrFlow Profile & Co Custom Objects"
                " and writes them to an Hrflow.ai source"
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadProfileActionParameters", format=format_into_hrflow_profile
            ),
            origin=SalesforceProfileWarehouse,
            target=HrFlowProfileWarehouse,
            action_type=ActionType.inbound,
        ),
        ConnectorAction(
            name=ActionName.push_profile,
            trigger_type=WorkflowType.catch,
            description=(
                "Pushs specific Profile from HrFlow and writes"
                " it to HrFlow_Profile__c & Co Custom Objects in Salesforce"
            ),
            parameters=BaseActionParameters.with_defaults(
                "PushProfileActionParameters", format=format_into_salesforce_profile
            ),
            origin=HrFlowProfileWarehouse,
            target=SalesforceProfileWarehouse,
            action_type=ActionType.outbound,
        ),
        ConnectorAction(
            name=ActionName.pull_job_list,
            trigger_type=WorkflowType.pull,
            description=(
                "Retrieves jobs from Salesforce HrFlow Job Custom Object"
                " and writes them to an Hrflow.ai board"
            ),
            parameters=BaseActionParameters.with_defaults(
                "ReadJobActionParameters", format=format_job
            ),
            origin=SalesforceJobWarehouse,
            target=HrFlowJobWarehouse,
            action_type=ActionType.inbound,
        ),
    ],
)
